# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Low level conversions, e.g. string to time.

Transformation of application data to a formal text representation according
to an associated :class:`~.Portable` type, and back again. These text
fragments become the *on-the-wire* representation for data such as times, that
might otherwise end up as unreadable integer or floating point numbers. These
transformations guarantee correctness of the encode-decode process, i.e. the
original application value will be recovered in the decoding process. System
limitations, such as floating point representation, mean that the guarantee
cannot be extended to related encode-decode processes.
"""

# .. autofunction:: world_to_text
# .. autofunction:: clock_to_text

__docformat__ = 'restructuredtext'

import time
import datetime
from enum import Enum
import re
import uuid

from .virtual_memory import *
from .convert_signature import *
from .convert_type import *

__all__ = [
	'ConversionError',
	'ConversionEncodeError',
	'ConversionDecodeError',

	'clock_to_text',
	'text_to_clock',
	'span_break',
	'span_to_text',
	'text_to_span',
	'UTC',
	'world_to_text',
	'text_to_world',
	'delta_to_text',
	'text_to_delta',
	'uuid_to_text',
	'text_to_uuid',
	'type_to_text',
	'text_to_type',
	'clock_now',
	'clock_at',
	'clock_break',
	'clock_span',
	'world_now',
	'world_at',
	'world_break',
	'world_delta',
]

# Exceptions
#
class ConversionError(Exception):
	"""Base exception for all conversion exceptions."""

class ConversionDecodeError(ConversionError):
	"""Exception raised on failure of one of the decoding functions."""

	def __init__(self, need: str, text: str):
		"""Construct the exception.

		:param need: type to be decoded
		:param text: description of failure
		"""
		self.need = need
		self.text = text

	def __str__(self):
		"""Auto-convert to str."""
		return 'cannot recover %s from "%s"' % (
			self.need, self.text)

class ConversionEncodeError(ConversionError):
	"""Exception raised on failure of one of the encoding functions."""

	def __init__(self, need: str, text: str):
		"""Construct the exception.

		:param need: type to be encoded
		:param text: description of failure
		"""
		self.need = need
		self.text = text

	def __str__(self):
		"""Auto-convert to str."""
		return 'cannot represent %s as a %s' % (
			self.text, self.need)

# Specification of time formats.
ISO_CLOCK_RE = r'^([0-9]{1,4})-([0-9]{1,2})-([0-9]{1,2})[tT]([0-9]{1,2}):([0-9]{1,2}):([0-9]{1,2})(\.([0-9]+))?$'
iso_clock = re.compile(ISO_CLOCK_RE)

ISO_8601_RE = r'^([0-9]{1,4})-([0-9]{1,2})-([0-9]{1,2})(T([0-9]{1,2})(:([0-9]{1,2})(:([0-9]{1,2})(\.([0-9]{1,6}))?)?)?([-+]([0-9]{1,2}):?([0-9]{1,2})(\.([0-9]{1,6}))?)?)?$'
iso_8601 = re.compile(ISO_8601_RE)

def clock_to_text(t):
	"""Convert host time value to standard clock-on-the-wall time representation. Return string."""
	a = '%f' % (t,)
	d = a.find('.')
	if d == -1:
		raise ConversionEncodeError('ClockTime', a)
	h = a[:d]
	t = a[d + 1:]
	if not h.isdigit() or not t.isdigit():
		# NOTE: catches times before epoch (negatives)
		raise ConversionEncodeError('ClockTime', a)
	i = int(h)
	f = int(t)
	t9 = time.localtime(i)
	iso = time.strftime(ISO_8601_NO_SECONDS, t9)
	s = t9[5]
	if f == 0:
		return '%s:%02d' % (iso, s)
	z = t.rstrip('0')
	return '%s:%02d.%s' % (iso, s, z)

def group_or_zero(s):
	if s is None:
		return 0
	return int(s)

def text_to_clock(s):
	"""Convert a standard clock-on-the-wall time string to a host time value. Return float."""
	m = iso_clock.match(s)
	if not m or s[-1] == 'Z':
		raise ConversionDecodeError('ClockTime', s)
	t9 = (group_or_zero(m.group(1)),
		group_or_zero(m.group(2)),
		group_or_zero(m.group(3)),
		group_or_zero(m.group(4)),
		group_or_zero(m.group(5)),
		group_or_zero(m.group(6)),
		0,
		0,
		-1)
	# Local time, no zone offset and with any
	# fractional part added in.
	c = time.mktime(t9)
	f = m.group(7)
	if f is not None:
		f = float(f)
		if c < 0.0:
			return c - f
		return c + f
	return c

# Helpers for time span values.
#
SPAN_TEXT = r'^[-]?(([0-9]+)d)?(([0-9]+)h)?(([0-9]{1,2})m)?(([0-9]{1,2}(\.[0-9]+)?)s)?$'
spanner = re.compile(SPAN_TEXT)

SECONDS_PER_MINUTE = 60
SECONDS_PER_HOUR = 60 * SECONDS_PER_MINUTE
SECONDS_PER_DAY = 24 * SECONDS_PER_HOUR

def span_break(s):
	"""Break down a timespan value into standard quanta. Return tuple."""
	n = s < 0.0
	if n:
		s = -s
	i = int(s)	  # Whole part.
	f = s % 1	   # Fractional part.

	D = int(i / SECONDS_PER_DAY)	 # Grab largest sub-value
	i %= SECONDS_PER_DAY		# Then discard same.

	H = int(i / SECONDS_PER_HOUR)
	i %= SECONDS_PER_HOUR

	M = int(i / SECONDS_PER_MINUTE)
	S = i % SECONDS_PER_MINUTE

	return n, D, H, M, S, f

def span_to_text(s):
	"""Convert a host timespan value to a standard text representation. Return string."""
	n, D, H, M, S, f = span_break(s)

	r = ''
	if n: r += '-'
	if D: r += '%dd' % (D,)
	if H: r += '%dh' % (H,)
	if M: r += '%dm' % (M,)
	if S:
		if f == 0.0:
			r += '%ds' % (S,)
		else:
			t = '%f' % (S + f,)
			r += t.rstrip('0')
			r += 's'
	else:
		if f != 0.0:
			t = '%f' % (S + f,)
			r += t.rstrip('0')
			r += 's'

	if len(r):
		return r
	return '0s'

def text_to_span(t):
	"""Convert a standard text representation of a timespan to a host value. Return float."""
	if len(t) == 0:
		raise ConversionDecodeError('TimeSpan', t)
	# Regular expression and subsequent processing allows for clean redundancy
	# in text representation of timespan, e.g. "1h" is the same as "1h0m0s".
	m = spanner.match(t)
	if not m:
		raise ConversionDecodeError('TimeSpan', t)
	n = t[0] == '-'
	D = group_or_zero(m.group(2))
	H = group_or_zero(m.group(4))
	M = group_or_zero(m.group(6))
	s = D * SECONDS_PER_DAY + H * SECONDS_PER_HOUR + M * SECONDS_PER_MINUTE

	if m.group(7):
		if m.group(9):
			s += float(m.group(8))
		else:
			s += float(int(m.group(8)))
	else:
		s = float(s)
	if n:
		return -s
	return s

ISO_8601_NO_SECONDS = '%Y-%m-%dT%H:%M'

#
#
UTC = datetime.timezone.utc

# Very sensitive functions in the area of fractional
# part and time conversion functions, e.g. gmtime and
# timegm.
def world_to_text(dt):
	"""Convert datetime value to standard ISO 8601 time representation. Return string."""
	tz = dt.tzinfo
	if tz is None:
		cannot = 'cannot represent a naive datetime'
		raise ValueError(cannot)
	if not isinstance(tz, datetime.timezone):
		cannot = 'cannot represent a tzinfo that is not a datetime.timezone'
		raise ValueError(cannot)

	ymdhms = dt.strftime('%Y-%m-%dT%H:%M:%S')

	f = dt.microsecond == 0
	if f:
		fraction = ''
	else:
		fraction = dt.strftime('.%f')
		fraction = fraction.rstrip('0')

	if tz == UTC:
		zone = ''
	else:
		zone = dt.strftime('%z')

	s = '%s%s%s' % (ymdhms, fraction, zone)
	return s

DIGITS_IN_MICRO = 6

def text_to_world(s):
	"""Convert a standard ISO 8601 string to an application value. Return datetime."""
	m = iso_8601.match(s)
	if not m:
		raise ConversionDecodeError('WorldTime', s)

	# Date parts. Required.
	dY = group_or_zero(m.group(1))
	dM = group_or_zero(m.group(2))
	dD = group_or_zero(m.group(3))

	# Time parts.
	tH = tM = tS = 0
	tF = 0

	# Zone parts.
	zH = zM = zS = 0
	zF = 0
	tz = UTC

	x = m.group(4)  # Is there a time and optional zone?
	if x:
		tH = group_or_zero(m.group(5))
		x = m.group(6)  # Is there a minutes, seconds and optional fraction?
		if x:
			tM = group_or_zero(m.group(7))
			x = m.group(8)  # Is there a seconds and optional fraction?
			if x:
				tS = group_or_zero(m.group(9))
				x = m.group(10) # Is there a fraction?
				if x:
					tF = m.group(11) # Fractional digits
					tN = len(tF)
					# Standard expects 6 fractional digits but regex allows from 1 to 6.
					# Unrealistic to expect editing that always conforms to 6 digits.
					# But still, technically this can never happen.
					if tN > DIGITS_IN_MICRO:
						raise ConversionDecodeError('WorldTime', s)
					tF = int(tF)
					# Adjust for the custom strip of trailing
					# zeroes.
					if tF and tN < DIGITS_IN_MICRO:
						tF = tF * 10 ** (DIGITS_IN_MICRO - tN)
		x = m.group(12)	 # Is a timezone present?
		if x:
			zH = group_or_zero(m.group(13))
			zM = group_or_zero(m.group(14))
			x = m.group(15)				 # Is there seconds and optional fraction?
			if x:
				zS = group_or_zero(m.group(16))  # Seconds
				x = m.group(17)				 # Is there a fraction?
				if x:
					zF = m.group(18)  # Fraction digits, no dot.
					zN = len(zF)
					# See same handling of fraction above.
					if zN > DIGITS_IN_MICRO:
						raise ConversionDecodeError('WorldTime', s)
					zF = int(zF)
					# Adjust for the custom strip of trailing
					# zeroes.
					if zF and zN < DIGITS_IN_MICRO:
						zF = zF * 10 ** (DIGITS_IN_MICRO - zN)

			td = datetime.timedelta(hours=zH, minutes=zM, seconds=zS, microseconds=zF)
			tz = datetime.timezone(td)

	dt = datetime.datetime(dY, dM, dD,
		hour=tH, minute=tM, second=tS,
		microsecond=tF,
		tzinfo=tz)
	return dt

# Helpers for timedelta values.
#

def delta_to_text(d):
	"""Convert a TimeDelta value to its text representation. Return string."""
	t = d.total_seconds()
	n, D, H, M, S, f = span_break(t)

	dash = ''
	if n:
		dash = '-'

	if f == 0.0:
		S = '%02d' % (S,)
	else:
		S = '%02f' % (S + f,)
		S = S.rstrip('0')

	if D == 0:
		return '%s%02d:%02d:%s' % (dash, H, M, S)
	return '%s%d:%02d:%02d:%s' % (dash, D, H, M, S)

# Helpers for time delta values.
#
DELTA_TEXT = r'^-?(([0-9]+):)?([0-9]{1,2}):([0-9]{1,2}):([0-9]{1,2})(\.([0-9]+))?$'
delta = re.compile(DELTA_TEXT)

def text_to_delta(t):
	"""Convert a standard text representation of a time delta to a host value. Return timedelta."""
	m = delta.match(t)
	if not m:
		raise ConversionDecodeError('TimeDelta', t)
	n = t[0] == '-'
	D = group_or_zero(m.group(2))
	H = int(m.group(3))
	M = int(m.group(4))
	S = int(m.group(5))

	F = m.group(7)
	if F is None:
		F = 0
	else:
		x = len(F)
		if x > DIGITS_IN_MICRO:
			raise ConversionDecodeError('TimeDelta', t)
		F = int(F)
		# Adjust for the custom strip of trailing
		# zeroes.
		if F and x < DIGITS_IN_MICRO:
			F = F * 10 ** (DIGITS_IN_MICRO - x)

	if n:
		D = -D
		H = -H
		M = -M
		S = -S
		F = -F

	dt = datetime.timedelta(days=D,
		hours=H,
		minutes=M,
		seconds=S,
		microseconds=F)
	return dt

#
#
def uuid_to_text(u):
	"""Convert a host UUID value to a standard text representation. Return string."""
	t = str(u)
	return t

def text_to_uuid(t):
	"""Convert a standard text representation to a host UUID value. Return UUID."""
	u = uuid.UUID(t)
	return u

#
#
def text_to_type(t):
	"""Convert the dotted string *s* to a class, or None."""
	s = lookup_signature(t)
	return s

def type_to_text(s):
	"""Convert the class *c* to a dotted string representation."""
	t = portable_to_signature(s)
	return t

#
#
def clock_now():
	"""Query the platform for the current ClockTime epoch. Return float."""
	t = time.time()
	return t

def clock_at(year, month, day, hour=0, minute=0, second=0, millisecond=0, microsecond=0):
	"""Convert the date and time arguments into a ClockTime epoch. Return float."""
	m = millisecond / 10 ** 3
	c = microsecond / 10 ** 6
	s = (year, month, day, hour, minute, second, 0, 0, -1)
	t = time.mktime(s)
	t = t + m + c
	return t

def clock_break(t):
	"""Break down a ClockTime value into standard quanta. Return a 10-tuple.

	A microseconds value is appended to the standard time.struct_time 9-tuple. It
	is taken from the fractional part of the value returned by time epoch.

	[0] Year
	[1] Month
	[2] Day of month
	[3] Hour
	[4] Minute
	[5] Second
	[6] Day of week
	[7] Day of year
	[8] Is daylight saving
	[9] Microseconds

	The standard time.localtime function turns a float into Gregorian calendar
	information. Unfortunately it loses the fractional part of a time epoch along
	the way, making format conversions a lossy process. This is an implementation
	that preserves the fractional part but produces a tuple that cannot be
	presented to time.mktime. This function is retained for special uses and
	as a record.
	"""
	i = int(t)
	f = t % 1
	t9 = time.localtime(i)
	if t9.tm_isdst:
		_ = 1
	Y = t9[0]
	M = t9[1]
	D = t9[2]
	H = t9[3]
	m = t9[4]
	S = t9[5]
	w = t9[6]
	y = t9[7]
	s = t9[8]
	X = int(f * 10 ** 6)
	return Y, M, D, H, m, S, w, y, s, X

def clock_span(days=0, hours=0, minutes=0, seconds=0, milliseconds=0, microseconds=0):
	"""Convert standard time quanta to a host time value. Return float."""
	m = milliseconds / 10 ** 3
	c = microseconds / 10 ** 6
	s = days * SECONDS_PER_DAY + hours * SECONDS_PER_HOUR + minutes * SECONDS_PER_MINUTE + seconds
	r = float(s) + m + c
	return r

def world_now(tz=None):
	"""Query the platform for the current WorldTime. Return datetime."""
	tz = tz or UTC
	t = datetime.datetime.now(tz=tz)
	return t

def world_at(year, month, day, hour=0, minute=0, second=0, millisecond=0, microsecond=0, tz=None):
	"""Convert the date, time and timezone arguments into a WorldTime. Return datetime."""
	p = millisecond * 10 ** 3
	m = microsecond + p
	if tz is None:
		t = datetime.datetime(year, month, day, hour, minute, second, m, tzinfo=UTC)
		return t

	t = datetime.datetime(year, month, day, hour, minute, second, m, tz)
	t = t.astimezone(UTC)
	return t

# Not really needed as datetime object provides
# direct access to breakdown. Included for
# symmetry.
def world_break(w, tz=None):
	"""Break down a WorldTime value into standard quanta. Return tuple."""
	if tz is not None:
		p = w.astimezone(tz)
	else:
		p = w
	Y = p.year
	M = p.month
	D = p.day
	H = p.hour
	m = p.minute
	S = p.second
	X = p.microsecond
	return Y, M, D, H, m, S, X

def world_delta(days=0, hours=0, minutes=0, seconds=0, milliseconds=0, microseconds=0):
	"""Convert the days, hours, minutes and seconds into a TimeSpan epoch difference. Return float."""
	return datetime.timedelta(days=days,
		hours=hours, minutes=minutes, seconds=seconds,
		milliseconds=milliseconds, microseconds=microseconds)



# object_startup_test.py
import kipjak as kj
import test_main

def main(self):
	t = kj.text_to_world('1963-03-26T02:24')
	a = self.create(kj.ProcessObject, test_main.main, b=32, c=99, t=t)
	m, i = self.select(kj.Returned, kj.Stop)
	if isinstance(m, kj.Returned):
		return m.message					# Return type of main must match test_main.main.
	self.send(m, a)
	self.select(kj.Returned)
	return kj.Aborted()

kj.bind(main, return_type=kj.Any())

if __name__ == '__main__':
	kj.create(main)



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implementation of the XML codec.

The plug-in part of the encoding/decoding stack and the default. Involves
two functions that transform between the internal, flattened memory
representation and the actual encoding. With a little luck there is not
much more to do than call the read/write, or load/save methods on the
encoding-specific parser module.

* word_to_xml - generate the JSON representation of an application word.
* xml_to_word - recovers an application word from a JSON representation.

Also refer to;

* https://www.w3.org/XML/
* https://en.wikipedia.org/wiki/XML
"""

__docformat__ = 'restructuredtext'

# .. autoclass:: CodecXml
# .. autofunction:: word_to_xml
# .. autofunction:: xml_to_word

import base64
import uuid
import types
import io

#
#
from defusedxml.ElementTree import fromstring, ParseError
from defusedxml import DTDForbidden, EntitiesForbidden, ExternalReferenceForbidden

from xml.dom import minidom

from .virtual_memory import *
from .convert_memory import *
from .virtual_runtime import *
from .message_memory import *
from .virtual_codec import *

__all__ = [
	'word_to_xml',
	'xml_to_word',
	'CodecXml'
]

# Code fragment per generic type, for output
# of pretty_format XML.

# Output a simple XML element - bool, int, float.
def name_value_pretty(c, tag, name, value, tabs):
	r = c.representation
	tab = c.find_tab(tabs)
	r.write(tab)
	r.write('<')
	r.write(tag)
	if name:
		r.write(' name="')
		r.write(name)
		r.write('"')
	if value:
		r.write(' value="')
		r.write(value)
		r.write('"')
	r.write('/>\n')

# Output of an ansar String or Unicode. Special XML
# characters need escaping and XML does not do
# non-ascii completely.
ampersand = {
	'&': '&amp;',
	'<': '&lt;',
	'>': '&gt;',
	'\'': '&apos;',
	'"': '&quot;',
}

# Printables according to
# XML specs.
xml_printing = set()

xml_printing.add(9)	# TAB
xml_printing.add(10)   # LF
xml_printing.add(13)   # CR

# From space to tilde, i.e. 127
# below is exclusive boundary.
for i in range(32, 127):
	xml_printing.add(i)

# Remove the tag/escape chars.
for k, v in ampersand.items():
	i = ord(k)
	xml_printing.discard(i)

def escaped_str(c, w):
	r = c.representation
	for c in w:
		i = ord(c)
		# 1. Printable
		# 2. Named character reference
		# 3. Numeric character reference
		if i in xml_printing:
			r.write(c)
			continue
		try:
			escaped = ampersand[c]
			r.write(escaped)
			continue
		except KeyError:
			pass
		numeric = '&#%d;' % (i,)
		r.write(numeric)

def w2t_pretty_bool(c, w, name, tabs):
	if w:
		name_value_pretty(c, 'boolean', name, 'true', tabs)
	else:
		name_value_pretty(c, 'boolean', name, 'false', tabs)

def w2t_pretty_int(c, w, name, tabs):
	value = '%d' % (w,)
	name_value_pretty(c, 'integer', name, value, tabs)

def w2t_pretty_float(c, w, name, tabs):
	value = '%f' % (w,)
	name_value_pretty(c, 'float', name, value, tabs)

def w2t_pretty_str(c, w, name, tabs):
	r = c.representation
	tab = c.find_tab(tabs)
	r.write(tab)
	r.write('<string')
	if name:
		r.write(' name="')
		r.write(name)
		r.write('"')
	r.write('>')
	escaped_str(c, w)
	# Tab here might be included in the recovered
	# value. Leave it dangling.
	r.write('</string>\n')

def w2t_pretty_list(c, w, name, tabs):
	r = c.representation
	tab = c.find_tab(tabs)
	r.write(tab)
	r.write('<list')
	if name:
		r.write(' name="')
		r.write(name)
		r.write('"')
	r.write('>\n')
	for a in w:
		word_to_text_pretty(c, a, None, tabs + 1)
	r.write(tab)
	r.write('</list>\n')

def w2t_pretty_dict(c, w, name, tabs):
	r = c.representation
	tab = c.find_tab(tabs)
	r.write(tab)
	r.write('<message')
	if name:
		r.write(' name="')
		r.write(name)
		r.write('"')
	r.write('>\n')
	for k, v in w.items():
		word_to_text_pretty(c, v, k, tabs + 1)
	r.write(tab)
	r.write('</message>\n')

def w2t_pretty_none(c, w, name, tabs):
	name_value_pretty(c, 'null', name, None, tabs)

# Map the generic type to a dedicated
# transform function.
w2t_pretty = {
	bool: w2t_pretty_bool,
	int: w2t_pretty_int,
	float: w2t_pretty_float,
	str: w2t_pretty_str,
	list: w2t_pretty_list,
	dict: w2t_pretty_dict,
	NoneType: w2t_pretty_none,
}

def word_to_text_pretty(c, w, name, tabs=0):
	try:
		a = getattr(w, '__class__')
	except AttributeError:
		raise TypeError('word is unusable')

	try:
		f = w2t_pretty[a]
	except KeyError:
		raise TypeError('no transformation for word %s' % (a.__name__,))

	# Apply the transform function
	return f(c, w, name, tabs)

# Code fragment per generic type, for output
# of non-pretty_format XML.

# Output a simple XML element - bool, int, float.
def name_value(c, tag, name, value):
	r = c.representation
	r.write('<')
	r.write(tag)
	if name:
		r.write(' name="')
		r.write(name)
		r.write('"')
	if value:
		r.write(' value="')
		r.write(value)
		r.write('"')
	r.write('/>')

def w2t_bool(c, w, name):
	if w:
		name_value(c, 'boolean', name, 'true')
	else:
		name_value(c, 'boolean', name, 'false')

def w2t_int(c, w, name):
	value = '%d' % (w,)
	name_value(c, 'integer', name, value)

def w2t_float(c, w, name):
	value = '%f' % (w,)
	name_value(c, 'float', name, value)

def w2t_str(c, w, name):
	r = c.representation
	r.write('<string')
	if name:
		r.write(' name="')
		r.write(name)
		r.write('"')
	r.write('>')
	escaped_str(c, w)
	# Tab here might be included in the recovered
	# value. Leave it dangling.
	r.write('</string>')

def w2t_list(c, w, name):
	r = c.representation
	r.write('<list')
	if name:
		r.write(' name="')
		r.write(name)
		r.write('"')
	r.write('>')
	for a in w:
		word_to_text(c, a, None)
	r.write('</list>')

def w2t_dict(c, w, name):
	r = c.representation
	r.write('<message')
	if name:
		r.write(' name="')
		r.write(name)
		r.write('"')
	r.write('>')
	for k, v in w.items():
		word_to_text(c, v, k)
	r.write('</message>')

def w2t_none(c, w, name):
	name_value(c, 'null', name, None)

w2t = {
	bool: w2t_bool,
	int: w2t_int,
	float: w2t_float,
	str: w2t_str,
	list: w2t_list,
	dict: w2t_dict,
	NoneType: w2t_none,
}

# Transform the generic data to rather inconvenient XML form.
def word_to_text(c, w, name):
	try:
		a = getattr(w, '__class__')
	except AttributeError:
		raise TypeError('word is unusable')

	try:
		f = w2t[a]
	except KeyError:
		raise TypeError('no transformation for word %s' % (a.__name__,))

	# Apply the transform function
	return f(c, w, name)

# Generate XML text representation of a
# generic word.
def word_to_xml(c, w):
	"""Generate the XML representation of a generic word.

	:param c: an active codec
	:type c: a Codec-based object
	:param w: a generic word
	:rtype: the XML text.
	"""
	c.representation = io.StringIO()
	if c.pretty_format:
		c.representation.write('<?xml version="1.0" ?>\n')
		word_to_text_pretty(c, w, None)
	else:
		c.representation.write('<?xml version="1.0" ?>')
		word_to_text(c, w, None)

	return c.representation.getvalue()

# Decoding - from parsing of JSON to transformation
# into app data items.

def xml_to_word(c, x):
	"""Produce a generic word from the parsing of a text XML representation.

	:param c: an active codec
	:type c: a Codec-based instance
	:param x: the XML text
	:type x: string
	:rtype: a generic word.
	"""
	try:
		e = fromstring(x, forbid_dtd=True, forbid_entities=True, forbid_external=True)
	except ParseError as t:
		raise ValueError(str(t))
	except DTDForbidden as t:
		raise ValueError(str(t))
	except EntitiesForbidden as t:
		raise ValueError(str(t))
	except ExternalReferenceForbidden as t:
		raise ValueError(str(t))

	w = element_to_word(c, e)
	return w

# From generic data (after parsing) to final python
# representation in the application.

def e2w_boolean(c, e):
	v = e.get('value')
	return v == 'true'

def e2w_integer(c, e):
	v = e.get('value')
	return int(v)

def e2w_float(c, e):
	v = e.get('value')
	return float(v)

def e2w_string(c, e):
	# An empty string is encoded as an empty XML element
	# and the parser considers that to be a None string.
	# If the original value was a None it would have been
	# encoded as a null element.
	if e.text is None:
		return str()
	return e.text

def e2w_list(c, e):
	v = [element_to_word(c, s) for s in e]
	return v

def e2w_message(c, e):
	v = {s.get('name'): element_to_word(c, s) for s in e}
	return v

def e2w_null(c, w):
	return None

#
#
e2w = {
	'boolean': e2w_boolean,
	'integer': e2w_integer,
	'float': e2w_float,
	'string': e2w_string,
	'list': e2w_list,
	'message': e2w_message,
	'null': e2w_null,
}

#
#
def element_to_word(c, e):
	a = e.tag	 # The XML element.

	try:
		f = e2w[a]
	except KeyError:
		raise TypeError('no transformation for XML tag "%s"' % (a,))

	return f(c, e)

# Define the wrapper around the JSON encoding
# primitives.
class CodecXml(Codec):
	"""Encoding and decoding of XML representations."""

	EXTENSION = 'xml'
	SINGLE_TAB = '  '

	def __init__(self, return_proxy=None, local_termination=None, pretty_format=False, decorate_names=True):
		"""Construct an XML codec."""
		Codec.__init__(self,
			CodecXml.EXTENSION,
			word_to_xml,
			xml_to_word,
			return_proxy, local_termination, pretty_format, decorate_names)
		self.representation = None
		self.tabstops = {}

	def find_tab(self, tabs):
		"""Generate the tab-space indicated by the tabs count. Return a string of spaces."""
		try:
			tab = self.tabstops[tabs]
		except KeyError:
			tab = CodecXml.SINGLE_TAB * tabs
			self.tabstops[tabs] = tab
		return tab



# test_library.py
import kipjak as kj

from test_api import *
from test_function import *


def library(self):
	'''Provide an API to the framework. Return nothing.'''

	# The process runs until it receives a Stop.
	while True:
		m = self.input()

		if isinstance(m, Xy):				# Expected request.
			pass
		elif isinstance(m, kj.Faulted):		# Any fault.
			return m
		elif isinstance(m, kj.Stop):		# Terminate, e.g. control-c.
			return kj.Aborted()
		else:
			continue

		# Process the framework request.
		# Commits the entire process to the function call.
		table = texture(self, x=m.x, y=m.y)
		self.send(kj.cast_to(table, table_type), self.return_address)

kj.bind(library, entry_point=[Xy,])		# Register with the framework. Declare the API.

if __name__ == '__main__':		# Process entry-point.
	kj.create(library)



# test_publisher_machine.py
import kipjak as kj

from test_api import *
from test_function import *

# A bare-bones, machine-based implementation of a published network service.
DEFAULT_NAME = 'acme'

class Publisher(kj.Point, kj.Stateless):
	def __init__(self, name: str=None):
		kj.Point.__init__(self)
		kj.Stateless.__init__(self)
		self.name = name or DEFAULT_NAME

def Publisher_Start(self, message):
	kj.publish(self, self.name)		# Register this object under the given name.

def Publisher_Xy(self, message):
	t = texture(self, x=message.x, y=message.y)
	m = kj.cast_to(t, table_type)
	self.reply(m)					# Respond to client.

def Publisher_Faulted(self, message):	# All faults routed here including
	self.complete(message)				# failure of publish().

def Publisher_Stop(self, message):
	self.complete(kj.Aborted())			# Leave all the housekeeping to the framework.

kj.bind(Publisher, (kj.Start, Xy, kj.Faulted, kj.Stop), entry_point=[Xy,])

if __name__ == '__main__':		# Process entry-point.
	kj.create(Publisher)



# test_server_1.py
import kipjak as kj
from test_api import Xy, table_type
from test_function_1 import texture

DEFAULT_ADDRESS = kj.HostPort('127.0.0.1', 5050)
SERVER_API = [Xy,]

def server(self, server_address: kj.HostPort=None):
	server_address = server_address or DEFAULT_ADDRESS

	kj.listen(self, server_address, http_server=SERVER_API)
	m = self.input()
	if not isinstance(m, kj.Listening):
		return m

	while True:
		m = self.input()
		if isinstance(m, Xy):
			pass
		elif isinstance(m, kj.Faulted):
			return m
		elif isinstance(m, kj.Stop):
			return kj.Aborted()
		else:
			continue

		response = texture(x=m.x, y=m.y)
		self.send(kj.cast_to(response, table_type), self.return_address)

kj.bind(server)

if __name__ == '__main__':
	kj.create(server)



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".

"""
__docformat__ = 'restructuredtext'

import re
from enum import Enum

from .virtual_memory import *
from .message_memory import *

__all__ = [
	'LOCAL_HOST',
	'HostPort',
	'equal_ipp',
	'LocalPort',
	'ScopeOfHost',
	'local_private_other',
]

#
LOCAL_HOST = '127.0.0.1'

class HostPort(object):
	"""Combination of an IP address or name, and a port number.

	:param host: IP address or name
	:param port: network port
	"""
	def __init__(self, host: str=None, port: int=None):
		self.host = host
		self.port = port

	def __str__(self):
		if self.host is None:
			return '(not set)'
		return f'{self.host}:{self.port}'

	def inet(self):
		return (self.host, self.port)

def equal_ipp(lhs, rhs):
	return lhs.host == rhs.host and lhs.port == rhs.port

class LocalPort(HostPort):
	def __init__(self, port: int=None):
		HostPort.__init__(self, LOCAL_HOST, port)

bind_message(HostPort)
bind_message(LocalPort, host=Unicode())

#
#
DOTTED_IP = re.compile(r'(\d+)\.(\d+)\.(\d+)\.(\d+)')

class ScopeOfHost(Enum):
	LOCAL=1
	PRIVATE=2
	OTHER=3

def local_private_other(ip):
	if ip is None:
		return ScopeOfHost.OTHER
	m = DOTTED_IP.match(ip)
	if m is None:
		return ScopeOfHost.OTHER
	# Have complete verification of dotted layout
	b0 = int(m.groups()[0])
	b1 = int(m.groups()[1])

	# Not dotted -------- None
	# 127.x.x.x --------- 0, localhost
	# 10.x.x.x ---------- 1, private
	# 192.168.x.x ------- 1, private
	# 172.[16-31].x.x --- 1, private
	# else -------------- 2, public

	if b0 == 127:
		return ScopeOfHost.LOCAL
	elif b0 == 10:
		return ScopeOfHost.PRIVATE
	elif b0 == 192 and b1 == 168:
		return ScopeOfHost.PRIVATE
	elif b0 == 172 and (b1 > 15 and b1 < 32):
		return ScopeOfHost.PRIVATE
	return ScopeOfHost.OTHER




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'LEFT NAME NUMBER RIGHT SEPARATORstatement : typetype : NAMEtype : NAME LEFT list RIGHTlist : arglist : list SEPARATOR argarg : NUMBERarg : type'

_lr_action_items = {'NAME':([0,4,10,],[3,3,3,]),'$end':([1,2,3,9,],[0,-1,-2,-3,]),'RIGHT':([3,5,6,7,8,9,11,],[-2,9,-4,-6,-7,-3,-5,]),'SEPARATOR':([3,5,6,7,8,9,11,],[-2,10,-4,-6,-7,-3,-5,]),'LEFT':([3,],[4,]),'NUMBER':([4,10,],[7,7,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'statement':([0,],[1,]),'type':([0,4,10,],[2,8,8,]),'list':([4,],[5,]),'arg':([4,10,],[6,11,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> statement","S'",1,None,None,None),
  ('statement -> type','statement',1,'p_statement_type','convert_type.py',146),
  ('type -> NAME','type',1,'p_type_name','convert_type.py',150),
  ('type -> NAME LEFT list RIGHT','type',4,'p_type_name_list','convert_type.py',162),
  ('list -> arg','list',1,'p_list_arg','convert_type.py',173),
  ('list -> list SEPARATOR arg','list',3,'p_list_list_arg','convert_type.py',177),
  ('arg -> NUMBER','arg',1,'p_arg_NUMBER','convert_type.py',184),
  ('arg -> type','arg',1,'p_arg_type','convert_type.py',189),
]



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Definition of the fundamental async object.

Async objects are created and then send messages to each other. There are
also timers to implement, and automated logging. There is also assistance
for managing async scenarios.
"""

__docformat__ = 'restructuredtext'

import os
import inspect
from copy import deepcopy
import types
import typing
from collections import deque
from queue import Empty
from time import time
from .general_purpose import *
from .virtual_memory import *
from .convert_memory import *
from .message_memory import *
from .convert_signature import *
from .convert_type import *
from .virtual_runtime import *
from .point_runtime import *
from .running_routine import *
from .routine_point import *
from .object_space import *
from .message_pump import *


__all__ = [
	'VP',
	'SelectTimer',
	'Point',
	'T1', 'T2', 'T3', 'T4',
	'StartTimer',
	'CancelTimer',
	'PointLog',
	'OnReturned',
	'returned_object',
	'Dispatching',
	'Buffering',
	'Threaded',
	'Channel',
	'Machine',
	'threaded_object',
	'object_dispatch',
	'no_ending',
	'halt',
]

# Point Runtime Addresses
# Initialized during startup or default
# to null, i.e. not available or no effect.
VP = Gas(log_address=NO_SUCH_ADDRESS,
	timer_address=NO_SUCH_ADDRESS,
	test_address=NO_SUCH_ADDRESS,
	circuit_address=NO_SUCH_ADDRESS,
	thread_classes={})

# Timing facility, i.e. Point.start().
class T1(object):
	"""Pre-declared timer class.

	A class suitable for passing to :meth:`~.Point.start`. The library
	provides the T1, T2, T3 and T4 timer classes for general use.
	"""
	pass

class T2(object): pass
class T3(object): pass
class T4(object): pass

bind_message(T1, copy_before_sending=False)
bind_message(T2, copy_before_sending=False)
bind_message(T3, copy_before_sending=False)
bind_message(T4, copy_before_sending=False)

class StartTimer(object):
	"""Message from point to clock service, requesting a timer."""
	def __init__(self, timer=None, seconds=1, repeating=False):
		self.timer = timer
		self.seconds = seconds
		self.repeating = repeating

class CancelTimer(object):
	"""Message from point to clock service, canceling a pending timer."""
	def __init__(self, timer=None):
		self.timer = timer

bind_message(StartTimer, copy_before_sending=False,
	timer=Type(),
	seconds=Float8(),
	repeating=Boolean(),
)
bind_message(CancelTimer, copy_before_sending=False, timer=Type())

# Logging facility, i.e. Point.log(), Point.console().
class PointLog(object):
	"""Object that records a moment in time and other details.

	:param stamp: the moment the log was created
	:type stamp: epoch float
	:param tag: an enumeration of the log level
	:type tag: a single-character string
	:param address: address of the async object
	:type address: tuple of int
	:param name: name of the class or function.
	:type name: str
	:param state: name of the current FSM state or None
	:type state: str
	:param text: free format text
	:type text: str
	"""
	def __init__(self, stamp=0.0, tag=None, address=None, name=None, state=None, text=None):
		self.stamp = stamp
		self.tag = tag
		self.address = address or NO_SUCH_ADDRESS
		self.name = name
		self.state = state
		self.text = text

bind_message(PointLog, copy_before_sending=False,
	message_trail=False, execution_trace=False,
	stamp=Float8(),
	tag=Enumeration(USER_TAG),
	address=Address(),
	name=Unicode(),
	state=Unicode(),
	text=Unicode(),
)

# Async unit test facility, i.e. Point.test()
class PointTest(object):
	"""Results of a Point.test().

	Captures the details of a check on runtime values.

	:param stamp: the moment the test was performed
	:type stamp: epoch float
	:param name: name of the class or function
	:type name: str
	:param condition: pass or fail
	:type condition: bool
	:param source: name of the module containing the test
	:type source: str
	:param line: line in the module
	:type line: int
	:param text: free format text
	:type text: str
	"""
	def __init__(self, stamp=None, name=None, condition=None, source=None, line=None, text=None):
		self.stamp = stamp
		self.name = name
		self.condition = condition
		self.source = source
		self.line = line
		self.text = text

bind_message(PointTest, copy_before_sending=False,
	message_trail=False, execution_trace=False,
	stamp=WorldTime(),
	name=Unicode(),
	condition=Boolean(),
	source=Unicode(),
	line=Integer8(),
	text=Unicode(),
)

def check_line():
	s = inspect.stack()[2]
	sf = s[1]	# Source file.
	ln = s[2]	# Line number.
	fn = s[3]	# Function name.
	if sf.find('tmp') == 1:
		_, sf = os.path.split(sf)
	if '<module>' in fn:
		fn = None
	return sf, ln, fn

# Low level control over behaviour of terminating
# object, i.e. some go quietly.
def returned_object(message, parent, address):
	send_a_message(Returned(message), parent, address)

# Automation of response to completion.
class OnReturned(object):
	"""Capture values needed for response to object completion.

	:param routine: type to be created
	:type routine: function or Point-based class
	:param args: positional parameters
	:type args: tuple
	:param kw: named parameters
	:type kw: dict
	"""
	def __init__(self, routine, args):
		self.routine = routine
		self.args = args

	def __call__(self, point, returned):
		m, p, a = un_cast(returned.message)
		point.returned_type = p
		return self.routine(point, m, self.args)	# Make the call.

#
#
class Point(object):
	"""The base class to all class-based asynchronous objects.

	Methods of this class are the user entry-point for SDL primitives such
	as;

	* :meth:`~.Point.create`
	* :meth:`~.Point.send`
	* :meth:`~.Point.reply`
	* :meth:`~.Point.forward`
	* :meth:`~.Point.start`
	* :meth:`~.Point.cancel`
	"""
	def __init__(self):
		self.object_address = NO_SUCH_ADDRESS	# Identity of this object.
		self.queue_address = NO_SUCH_ADDRESS	# Where are messages processed.
		self.parent_address = NO_SUCH_ADDRESS	# Who created this object.
		self.to_address = NO_SUCH_ADDRESS		# Delivery address on current message.
		self.return_address = NO_SUCH_ADDRESS	# Who sent the current message.
		self.assigned_queue = None				# Parent queue for non-threaded machine.
		self.object_ending = None

		self.returned_type = None
		self.received_type = None
		self.current_state = None

		self.address_context = {}
		self.aborted_message = None

	def create(self, object_type, *args, object_ending=returned_object, **kw):
		"""
		Create a new asynchronous object. Return the address.

		:param object_type: type of object to instantiate
		:type object_type: :ref:`object type<kj-object-type>`
		:param args: positional arguments passed to the new object
		:param object_ending: override standard termination
		:type object_ending: function
		:param kw: named arguments passed to the new object
		:rtype: Address
		"""
		return create_a_point(object_type, object_ending, self.object_address, args, kw)

	def send(self, message: Any, to: Address):
		"""Transfer a message to the specified address.

		:param message: message to be sent
		:param to: intended receiver of the message
		"""
		pf = self.__art__
		try:
			mf = getattr(message, '__art__')
		except AttributeError:
			mf = None
		if mf:
			xf = message.timer.__art__ if isinstance(message, (StartTimer, CancelTimer)) else mf
			if pf.message_trail and xf.message_trail:
				self.log(USER_TAG.SENT, 'Sent %s to <%08x>' % (mf.name, to[-1]))
			if mf.copy_before_sending:
				c = deepcopy(message)
				send_a_message(c, to, self.object_address)
				return
		send_a_message(message, to, self.object_address)

	def reply(self, message: Any):
		"""
		Send a response to the sender of the current message.

		This is a shorthand for ``self.send(m, self.return_address)``. Reduces
		keystrokes and risk of typos.

		:param message: the message to be sent
		"""
		self.send(message, self.return_address)

	def forward(self, message: Any, to: Address, return_address: Address):
		"""Send a message to an address, as if it came from a 3rd party.

		Send a message but override the return address with the address of
		another arbitrary object. To the receiver the message appears to
		have come from the arbitrary object.

		Useful when building relationships between objects. This allows objects
		to "drop out" of 3-way conversations, leaving simpler and faster 2-way
		conversations behind.

		:param message: the message to be sent
		:param to: intended receiver of the message
		:param return_address: the other object
		"""
		pf = self.__art__
		try:
			mf = getattr(message, '__art__')
			xf = message.timer.__art__ if isinstance(message, (StartTimer, CancelTimer)) else mf
		except AttributeError:
			mf = None

		if mf:
			if pf.message_trail and xf.message_trail:
				self.log(USER_TAG.SENT, 'Forward %s to <%08x> (from <%08x>)' % (mf.name, to[-1], return_address[-1]))
			if mf.copy_before_sending:
				c = deepcopy(message)
				send_a_message(c, to, return_address)
				return
		send_a_message(message, to, return_address)

	def start(self, timer, seconds: float, repeating: bool=False):
		"""Start the specified timer for this object.

		An instance of the ``timer`` class will be sent to this address after the
		given number of seconds. Any registered message can be used as a timer.
		kipjak provides the standard timers T1, T2, T3, and T4 for convenience
		and to reduce duplication.

		Timers are private to each async object and there is no limit to the
		number of pending timers an object may have. Starting a timer with the
		same class is not an error, the timeout for that timer is reset to the
		new number of seconds. It is also not an error to terminate an object
		with outstanding timers - expiry messages fall on the floor.

		It is difficult to make guarantees about the order that messages will
		arrive at an object. In the case of timers, its possible to receive
		the timeout after the sending of a cancellation. Async objects are best
		written to cope with every possible receive ordering.

		:param timer: type of the object that will be sent back on timeout
		:type timer: class
		:param seconds: time span before expiry
		:param repeating: enable endless repeat
		"""
		self.send(StartTimer(timer, seconds, repeating), VP.timer_address)

	def cancel(self, timer):
		"""Abort the specified timer for this object.

		Discard the pending timer. It is not an error to find that there is
		no such pending timer. The timeout can still be received after a
		cancellation.

		:param timer: the pending timer
		:type timer: class
		"""
		self.send(CancelTimer(timer), VP.timer_address)

	def complete(self, message: Any=None):
		"""Cause an immediate termination. The method never returns.

		:param message: message to be returned to parent.
		"""
		message = self.aborted_message or message
		raise Completion(message)

	def assign(self, address: Address, context=True):
		"""The specified child object is associated with the specified context, e.g. callback.

		:param address: address of the child object
		:param context: what the child object is doing on behalf of this object
		:type context: any
		"""
		self.address_context[address] = context

	def working(self):
		"""
		Check if there are child objects still active. Returns the count.

		:rtype: int
		"""
		return len(self.address_context)

	def progress(self, address: Address=None):
		"""Find the context associated with the specified child object. Return the context or None.

		If no address is provided the current return address is used.

		:param address: address of the child object
		:rtype: any
		"""
		a = address or self.return_address
		try:
			j = self.address_context[a]
		except KeyError:
			return None
		return j

	def running(self):
		"""Yield a sequence of context, address tuples."""
		for k, v in self.address_context.items():
			yield v, k

	def abort(self, message: Any=None):
		"""
		Initiate termination of all pending child objects. Return nothing.

		If required, a single value is stored for later retrieval
		as ``self.aborted_message``.

		:param message: value to save
		:rtype: None
		"""
		self.aborted_message = message
		for _, a in self.running():
			self.send(Stop(), a)
		n = len(self.address_context)
		return n

	def debrief(self, address: Address=None):
		"""
		Find the context associated with the child object. Return the context.

		If no address is provided the current return address is used.
		If a match is found the record is removed, decrementing the
		number of active contexts.

		:param address: address of the child object
		:rtype: any
		"""
		a = address or self.return_address
		c = self.address_context.pop(a, None)
		return c

	def on_return(self, address: Address, f, **kw):
		"""
		Register a callback, to be executed on termination of the specified object.

		:param address: address of the child object
		:param f: the function to be called
		:type f: function
		:param kw: named arguments to be saved and presented on the callback
		:type kw: dict
		"""
		c = OnReturned(f, Gas(**kw))
		self.assign(address, c)

	def continuation(self, address: Address, f, args: Gas):
		"""
		Register a chained callback, to be executed on termination of the specified object.

		:param address: address of the child object
		:param f: the function to be called
		:type f: function
		:param args: named arguments saved by a prior :meth:`~.Point.on_return`
		"""
		c = OnReturned(f, args)
		self.assign(address, c)

	def a_kv(self, a, kv):
		if len(a) > 0:
			note = ' '.join(a)
			if len(kv) > 0:
				eq = [f'{k}={v}' for k, v in kv.items()]
				t = ','.join(eq)
				text = f'{note} ({t})'
			else:
				text = note
		elif len(kv) > 0:
			eq = [f'{k}={v}'for k, v in kv.items()]
			text = ','.join(eq)
		else:
			text = None
		return text

	def log(self, tag, text):
		"""Generate a PointLog object at the specified level.

		This an internal function that should rarely be used directly
		by an application. Use debug(), trace(), etc instead.

		Forms a standard logging object and sends it to the logging
		service within the kipjak runtime.

		:param tag: one of the logging enumerations
		:type tag: USER_LOG
		:param text: the message to log
		:type text: str
		"""
		e = PointLog()
		e.stamp = time()
		e.tag = tag
		e.address = self.object_address
		e.name = self.__art__.name
		# Hack-ish implementation.
		# Should be common base and derived impl's.
		if self.current_state is None:
			e.state = None
		else:
			e.state = self.current_state.__name__

		e.text = text
		send_a_message(e, VP.log_address, self.object_address)

	def pass_fail(self, condition, source, line, text):
		p = PointTest()
		p.stamp = world_now()
		p.name = self.__art__.name

		p.condition = condition
		p.source = source
		p.line = line

		encoded = text.encode('utf-8')
		decoded = encoded.decode('ascii', 'backslashreplace')
		p.text = decoded
		send_a_message(p, VP.test_address, self.object_address)

	def debug(self, *a, **kv):
		"""Generate a log at level DEBUG.

		:param a: positional args catenated into a single string
		:type a: tuple
		:param kv: key-value pairs listed with a separating ``=`` character
		:type kv: dict
		"""
		if self.__art__.user_logs.value > USER_LOG.DEBUG.value:
			return
		text = self.a_kv(a, kv)
		if text:
			self.log(USER_LOG.DEBUG, text)

	def trace(self, *a, **kv):
		"""Generate a log at level TRACE.

		:param a: positional args catenated into a single string
		:type a: tuple
		:param kv: key-value pairs listed with a separating ``=`` character
		:type kv: dict
		"""
		if self.__art__.user_logs.value > USER_LOG.TRACE.value:
			return
		text = self.a_kv(a, kv)
		if text:
			self.log(USER_TAG.TRACE, text)

	def console(self, *a, **kv):
		"""Generate a log at level CONSOLE.

		:param a: positional args catenated into a single string
		:type a: tuple
		:param kv: key-value pairs listed with a separating ``=`` character
		:type kv: dict
		"""
		if self.__art__.user_logs.value > USER_LOG.CONSOLE.value:
			return

		text = self.a_kv(a, kv)
		if text:
			self.log(USER_TAG.CONSOLE, text)

	def sample(self, name, **kv):
		"""Generate a row in a sample stream.

		Logging is limited to a single positional ``name`` and a
		set of key-value pairs. This appears in the log with the same
		layout as ``console`` et al, but with the '&' object tag.
		The **kipjak** :ref:`log<command-reference-log>`
		command can be used to extract the series of samples and print
		them as a row of timestamped values separated by a tab. Logging
		sample values that may include ',' (comma) or '=' (equals) will
		eventually lead to unexpected output from the ``log`` command.

		:param name: name of this dataset
		:type name: str
		:param kv: key-value pairs listed with a separating ``=`` character
		:type kv: dict
		"""
		a = (name,)
		text = self.a_kv(a, kv)
		if text:
			self.log(USER_TAG.SAMPLE, text)

	def warning(self, *a, **kv):
		"""Generate a log at level WARNING.

		:param a: positional args catenated into a single string
		:type a: tuple
		:param kv: key-value pairs listed with a separating ``=`` character
		:type kv: dict
		"""
		if self.__art__.user_logs.value > USER_LOG.WARNING.value:
			return
		text = self.a_kv(a, kv)
		if text:
			self.log(USER_TAG.WARNING, text)

	def fault(self, *a, **kv):
		"""Generate a log at level FAULT.

		:param a: positional args catenated into a single string
		:type a: tuple
		:param kv: key-value pairs listed with a separating ``=`` character
		:type kv: dict
		"""
		if self.__art__.user_logs.value > USER_LOG.FAULT.value:
			return
		text = self.a_kv(a, kv)
		if text:
			self.log(USER_TAG.FAULT, text)

	def test(self, condition, note):
		"""Generate a log at level CHECK, dependent on the condition.

		A ``PointTest`` is also sent to an internal collection point, for
		later recovery, e.g. by test applications. This is sent for both
		pass and fail.

		:param condition: pass or fail
		:type condition: bool
		:param note: short description
		:type note: str
		"""
		s, l, _ = check_line()
		b = bool(condition)
		self.pass_fail(b, s, l, note)
		if b:
			return condition
		self.log(USER_TAG.CHECK, f'{note} ({s}:{l})')
		return condition

def halt(address):
	"""Mark the object at the specified address as halted.

	:param address: an async object
	:type address: kipjak address
	"""
	with OpenAddress(address) as c:
		if isinstance(c, Channel):
			c.halted = True

# Augmented input process, deriving straight
# from SDL.
MAXIMUM_REPLAYS = 8

class Player(object):
	"""Base for objects intending to operate message buffering with replay."""

	def __init__(self):
		"""Construct an instance of Player."""
		self.pending = deque()		# Recently saved.
		self.replaying = deque()	# Active replay
		self.get_frame = None

	def pull(self):
		"""Get the next replay message or a fresh message from the queue."""
		if self.replaying:
			mtr = self.replaying.popleft()
		else:
			mtr = self.get()
			if len(mtr) == 3:
				mtr.append(0)
			if self.pending:
				# Only replay those pending on the same address.
				other = deque()
				for p in self.pending:
					if p[1][-1] == mtr[1][-1]:
						self.replaying.append(p)
					else:
						other.append(p)
				self.pending = other
		self.get_frame = mtr
		return mtr[0], mtr[1], mtr[2]

	def pushback(self, message):
		"""Retain the [message, to, return] triplet for later replay."""
		mtr = self.get_frame
		# Cannot verify in the presence of constructed types.
		# This pushes back the most recent pull.
		#if id(message) != id(mtr[0]):
		#	return
		mtr[3] += 1
		if mtr[3] < MAXIMUM_REPLAYS:
			self.pending.append(mtr)

	def flush(self):
		try:
			while True:
				self.message_queue.get_nowait()
		except Empty:
			pass
		self.replaying.clear()
		self.pending.clear()

# Dedicated timer for managed messaging.
class SelectTimer(object):
	"""Timer for managed input."""
	pass

class Other(object):
	"""Capture un-declared input."""
	def __init__(self, message: Any=None):
		self.message = message

bind_message(SelectTimer)
bind_message(Other)

class Dispatching(Player):
	"""Provides input mechanism for any machine with its own queue."""

	def __init__(self):
		Player.__init__(self)
		"""Construct an instance of Dispatching."""

	def input(self):
		"""Block until the next message arrives. Return the value.

		Part of the "save" machinery that can buffer and replay messages. Sets addresses and
		type information (i.e. ``Point.received_type``) associated with the received message.

		:rtype: application value
		"""
		m, t, r = self.pull()
		self.to_address = t
		self.return_address = r
		m, p, a = un_cast(m)
		if self.__art__.execution_trace and (not a or a.execution_trace):
			tag = portable_to_tag(p)
			self.log(USER_TAG.RECEIVED, f'Received {tag} from <{r[-1]:08x}>')
		self.received_type = p
		return m

	def undo(self, message):
		"""Retain the [message, to, return] triplet, using values saved during input."""
		self.pushback(message)

class Buffering(Player):
	"""Base for any object intending to perform sophisticated I/O, i.e. channels and routines."""

	def __init__(self):
		Player.__init__(self)
		"""Construct an instance of Buffering."""

	def save(self, message):
		"""Pushback the most recent input message.

		Pass the application data that was returned by :meth:`~.Buffering.input`
		or :meth:`~.Buffering.select`.

		:param message: message to be saved
		"""
		self.pushback(message)

	def input(self):
		"""Block until the next message arrives. Return the value.

		Part of the "save" machinery that can buffer and replay messages. Sets addresses and
		type information (i.e. ``Point.received_type``) associated with the received message.

		:rtype: application value
		"""
		m, t, r = self.pull()
		self.to_address = t
		self.return_address = r
		m, p, a = un_cast(m)
		if self.__art__.execution_trace and (not a or a.execution_trace):
			tag = portable_to_tag(p)
			self.log(USER_TAG.RECEIVED, f'Received {tag} from <{r[-1]:08x}>')
		self.received_type = p
		return m

	def is_type(self, p: Portable):
		return self.received_type is p

	def select(self, *matching, saving: tuple=None, seconds: float=None):
		"""Expect one of the listed messages, with optional saving and timeout. Return a 2-tuple.

		Reads the input queue looking for a message matching one of the types
		listed in ``matching``. The method returns the application-ready data
		and the ordinal position of the match. Before dropping an unmatched
		message there is a check against the types listed in ``saving``. A match
		here results a "push back" of the message onto the input queue.

		For optimum performance the method also accepts the results of :func:`~.select_list`
		as replacements for ``matching`` and ``saving`` tuples.

		:param matching: message types to be accepted
		:type matching: the positional arguments tuple
		:param saving: message types to be deferred
		:type saving: tuple
		:param seconds: waiting period
		:type seconds: float
		:rtype: message object
		"""
		if len(matching) == 0:
			if seconds:
				matching += (Unknown, SelectTimer)
			else:
				matching = (Unknown,)
			matching = select_list_adhoc(*matching)
		elif isinstance(matching[0], SelectTable):
			matching = matching[0]
		else:
			matching = select_list_adhoc(*matching)

		if saving is None:
			pass
		elif isinstance(saving, SelectTable):
			pass
		elif isinstance(saving, tuple):
			saving = select_list_adhoc(*saving)
		else:
			saving = select_list_adhoc(saving)

		if seconds:
			matching += (SelectTimer,)
			self.start(SelectTimer, seconds)

		art = self.__art__
		while True:
			mtr = self.pull()
			m = mtr[0]
			self.to_address = mtr[1]
			self.return_address = mtr[2]
			a = self.return_address[-1]

			r = matching.find(m)
			if r is not None:
				if seconds:
					self.cancel(SelectTimer)
				if art.execution_trace:
					t = portable_to_tag(r[2])
					self.log(USER_TAG.RECEIVED, f'Received {t} from <{a:08x}>')
				self.received_type = r[2]
				return r[1], r[0]

			if saving and saving.find(m):
				self.save(m)
				continue

			if art.execution_trace:
				t = portable_to_tag(r[2])
				self.log(USER_TAG.RECEIVED, f'Dropped {t} from <{a:08x}>')

	def ask(self, q, r, a, saving=None, seconds=None):
		"""Query for a response while allowing reordering, with optional timer.


		:param q: query to be sent
		:type q: registered message
		:param r: response types to be detected and returned
		:type r: tuple
		:param a: async object to be queried
		:type a: kipjak address
		:param saving: response types to be detected and buffered
		:type saving: tuple
		:param seconds: waiting period
		:type seconds: float
		:rtype: message object
		"""
		self.send(q, a)
		if not isinstance(r, tuple):
			return self.select(r, saving=saving, seconds=seconds)
		return self.select(*r, saving=saving, seconds=seconds)

	def stop(self, a, r=(Returned,), saving=None, seconds=None):
		"""Request the termination of an object.

		:param a: async object to be queried
		:type a: kipjak address
		:param r: response types to be detected and returned
		:type r: tuple
		:param saving: response types to be detected and buffered
		:type saving: tuple
		:param seconds: waiting period
		:type seconds: float
		:rtype: message object
		"""
		return self.ask(Stop(), r, a, saving=saving, seconds=seconds)

class Threaded(Pump, Point, Dispatching):
	"""Base class for asynchronous machines that require dedicated threads.

	:param blocking: block on queue full
	:type blocking: bool
	:param maximum_size: number of pending messages
	:type maximum_size: int
	"""
	def __init__(self, blocking=False, maximum_size=PEAK_BEFORE_DROPPED):
		Pump.__init__(self, blocking=blocking, maximum_size=maximum_size)
		Point.__init__(self)
		Dispatching.__init__(self)

class Channel(Pump, Point, Buffering):
	"""A sync object.

	Used by sync code to access async features. An instance of a channel
	is created by kipjak on behalf of each "function" object, i.e. each instance
	of a function object gets a thread and a channel. The thread accesses
	async services through it private channel object. Also used by the
	OpenChannel context object.
	"""
	def __init__(self):
		Pump.__init__(self)
		Point.__init__(self)
		Buffering.__init__(self)
		self.halted = False

class Machine(object):
	"""Base for machines, providing for presentation of messages and save."""

	def __init__(self):
		"""Construct an instance of Machine."""

	def save(self, message):
		"""Retain the [message, to, return] triplet, i.e. call the parent queue."""
		self.assigned_queue.undo(message)

	def received(self, queue, message, return_address):
		"""A placeholder for erroneous use. Raises a ``PointConstructionError`` exception."""
		raise PointConstructionError('Message received by virtual base. Use Stateless or StateMachine.')

# Standard routines for running the
# message handlers for async class instances.
def threaded_object(queue):
	"""The thread object silently allocated to machines with dedicated threads."""
	a = queue.object_address
	queue.received(queue, Start(), queue.parent_address)
	while True:
		mtr = queue.pull()
		m = mtr[0]
		t = mtr[1]
		r = mtr[2]
		if t[-1] == a[-1]:
			queue.to_address = t
			queue.return_address = r
			queue.received(queue, m, r)

	# Termination occurs by raising of Completion exception
	# from within the received() method. The exception is
	# caught by the running_in_thread() function.

def object_dispatch(queue):
	"""The thread object that executes transitions for one or more machines."""
	a = queue.object_address
	while True:
		mtr = queue.pull()
		m = mtr[0]
		t = mtr[1]
		r = mtr[2]
		if t[-1] == a[-1]:
			if isinstance(m, Stop):	 # [1]
				return True
			continue

		p = find_object(t)
		if not p:
			continue

		try:
			p.to_address = t
			p.return_address = r
			p.received(queue, m, r)	 # [2]
			continue
		# Necessary replication of exceptions in
		# running_in_thread.
		except KeyboardInterrupt:
			s = 'unexpected keyboard interrrupt'
			p.fault(s)
			message = Faulted('object compromised', s)
		except SystemExit:
			s = 'unexpected system exit'
			p.fault(s)
			message = Faulted('object compromised', s)
		except Completion as c:
			message = c.message
		except Exception as e:
			s = str(e)
			s = f'unhandled exception ({s})'
			p.fault(s)
			message = Faulted('object faulted', s)
		except:
			s = 'unhandled opaque exception'
			p.fault(s)
			message = Faulted('object faulted', s)

		# Convert the exception to a message.
		if p.__art__.lifecycle:
			p.log(USER_TAG.DESTROYED, 'Destroyed')
		parent = p.parent_address

		return_type = p.__art__.return_type
		if return_type is None:
			pass
		elif isinstance(return_type, Any):
			pass
		elif isinstance(return_type, Portable):
			if not hasattr(message, '__art__'):
				message = (message, return_type)
		else:
			message = Faulted(f'unexpected return type for machine "{p.__class__.__name__}"')

		ending = p.object_ending
		destroy_an_object(t)
		ending(message, parent, t)

	# Termination of child objects occurs by raising of the
	# Completion exception from within the received() call [2].
	# The exception is caught and the termination protocol
	# concluded. Termination of the dispatcher [1] is the same
	# as for any object, by sending a Stop(). The running_in_thread
	# function concludes the protocol.

def no_ending(message, parent, address):
	pass

# Object creation.
#
def create_a_point(object_type, object_ending, parent_address, args, kw):
	if isinstance(object_type, types.FunctionType):
		return custom_routine(object_type, object_ending, parent_address, args, kw)
	elif issubclass(object_type, Point):
		if issubclass(object_type, Machine):
			if issubclass(object_type, Threaded):
				return object_and_thread(object_type, object_ending, parent_address, args, kw)
			else:
				return object_only(object_type, object_ending, parent_address, args, kw)
		if issubclass(object_type, Channel):
			return sync_object(object_type, parent_address, args, kw)
		else:
			raise PointConstructionError('Cannot create Point %r; not a machine or sync.' % (object_type,))
	else:
		raise PointConstructionError('Cannot create %r; not a function or Point.' % (object_type,))

def custom_routine(routine, object_ending, parent_address, args, kw):
	"""Create an async object around the supplied function.

	:param routine: the function to run within its own thread
	:type routine: function
	:param parent_address: object that called create()
	:type parent_address: kipjak address
	:param args: positional args to be forwarded to the function
	:type args: tuple
	:param kw: key-value args to be forwarded to the function
	:type kw: dict
	:rtype: kipjak address
	"""
	# A rude little adjustment to force the object over
	# to the input() system intended for dispatching.
	if routine == object_dispatch:
		object_type = Threaded   # Input and save.
	else:
		object_type = Channel	 # Input, select, ask and save.
	a, q = create_an_object(object_type, object_ending, parent_address, (), {})
	# Overlay the class runtime with the function runtime, i.e. at instance level.
	q.__art__ = routine.__art__
	if q.__art__.lifecycle:
		q.log(USER_TAG.CREATED, 'Created by <%08x>' % (parent_address[-1],))
	start_a_thread(q, routine, args, kw)
	return a

def object_and_thread(object_type, object_ending, parent_address, args, kw):
	"""Create an async object and thread around the supplied type.

	:param object_type: async object expecting its own thread
	:type object_type: registered class
	:param parent_address: object that called create()
	:type parent_address: kipjak address
	:param args: positional args to be forwarded to the function
	:type args: tuple
	:param kw: key-value args to be forwarded to the function
	:type kw: dict
	:rtype: kipjak address
	"""
	a, q = create_an_object(object_type, object_ending, parent_address, args, kw)
	# Assume the class context, i.e. a derived application class.
	if q.__art__.lifecycle:
		q.log(USER_TAG.CREATED, 'Created by <%08x>' % (parent_address[-1],))
	start_a_thread(q, threaded_object, (), {})
	# The threaded_object routine provides the Start message.
	return a

def object_only(object_type, object_ending, parent_address, args, kw):
	"""Create an async object of the supplied type.

	:param object_type: async object
	:type object_type: registered class
	:param parent_address: object that called create()
	:type parent_address: kipjak address
	:param args: positional args to be forwarded to the function
	:type args: tuple
	:param kw: key-value args to be forwarded to the function
	:type kw: dict
	:rtype: kipjak address
	"""
	a, q = create_an_object(object_type, object_ending, parent_address, args, kw)
	# Assume the object context.
	if q.__art__.lifecycle:
		q.log(USER_TAG.CREATED, 'Created by <%08x>' % (parent_address[-1],))
	send_a_message(Start(), a, parent_address)
	return a

def sync_object(object_type, parent_address, args, kw):
	"""Create a synchronous object in the midst of async objects.

	:param object_type: async object
	:type object_type: registered class
	:param parent_address: object that called create()
	:type parent_address: kipjak address
	:param args: positional args to be forwarded to the function
	:type args: tuple
	:param kw: key-value args to be forwarded to the function
	:type kw: dict
	:rtype: async object
	"""
	a, q = create_an_object(object_type, no_ending, parent_address, args, kw)
	# Assume the object context
	if q.__art__.lifecycle:
		q.log(USER_TAG.CREATED, 'Created by <%08x>' % (parent_address[-1],))
	return q



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Construction of message instances.

.. autofunction:: make
.. autofunction:: fake
"""

__docformat__ = 'restructuredtext'

import uuid
from datetime import MINYEAR, datetime, timedelta
from .virtual_memory import *
from .convert_memory import *
from .virtual_runtime import *
from .message_memory import *

__all__ = [
	'make',
	'fake',
]

# A mapping from kipjak to a constructor where
# no parameter is needed.
MAKE_CLASS = {
	Boolean: default_none,
	Byte: default_none,
	Character: default_none,
	Rune: default_none,
	Integer2: default_none,
	Integer4: default_none,
	Integer8: default_none,
	Unsigned2: default_none,
	Unsigned4: default_none,
	Unsigned8: default_none,
	Float4: default_none,
	Float8: default_none,
	Block: default_none,
	String: default_none,
	Unicode: default_none,
	ClockTime: default_none,
	TimeSpan: default_none,
	WorldTime: default_none,
	TimeDelta: default_none,
	UUID: default_none,
	Enumeration: default_none,
	Type: default_none,
	TargetAddress: default_none,
	Address: default_none,
	PointerTo: default_none,
	Any: default_none,
	VectorOf: default_vector,
	SetOf: default_set,
	MapOf: default_map,
	DequeOf: default_deque,
	# Need parameters;
	# ArrayOf
	# UserDefined
}

#
#
def make(t):
	"""Manufactures the Python equivalent of the memory description, or None."""
	if not is_portable(t):
		raise MessageRegistrationError(None, 'non-memory type presented for construction - %r' % (t,))

	try:
		c = MAKE_CLASS[t.__class__]
		return c()
	except KeyError:
		pass
	except AttributeError:
		raise MessageRegistrationError(None, 'internal failure to create from class')

	# Following types are more involved - cant be
	# ctor'd solely from the class.
	if isinstance(t, ArrayOf):
		d = [None] * t.size
		for i in range(t.size):
			d[i] = make(t.element)
		return d

	if isinstance(t, UserDefined):
		return t.element()

	raise MessageRegistrationError(None, 'internal failure to create from memory')

#
#
class Fake(object): pass

bind_message(Fake)

def character_bytes(): return b'c'
def rune_str(): return 'C'
def block_bytearray(): return bytearray([0x0c, 0x0a, 0x0f, 0x0e])
def string_bytes(): return b'CAFE'
def unicode_str(): return 'CAFE'

def fake_clock(): return datetime(1963, 3, 26).timestamp()
def fake_span(): return 0.5
def fake_world(): return datetime(1963, 3, 26)
def fake_delta(): return timedelta(seconds=0.5)
def fake_uuid(): return uuid.uuid4()
def fake_type_(): return Fake
def fake_target(): return [0x0c, 0x0a, 0x0f, 0x0e]
def fake_address(): return [0x0c, 0x0a, 0x0f, 0x0e]
def fake_any(): return [Fake, {}, []]

FAKE_CLASS = {
	Boolean: bool,
	Byte: int,
	Character: character_bytes,
	Rune: rune_str,
	Integer2: int,
	Integer4: int,
	Integer8: int,
	Unsigned2: int,
	Unsigned4: int,
	Unsigned8: int,
	Float4: float,
	Float8: float,
	Block: block_bytearray,
	String: string_bytes,
	Unicode: unicode_str,
	ClockTime: fake_clock,
	TimeSpan: fake_span,
	WorldTime: fake_world,
	TimeDelta: fake_delta,
	UUID: fake_uuid,
	Type: fake_type_,
	TargetAddress: fake_target,
	Address: fake_address,
	Any: fake_any,
}

def fake(t):
	"""Synthesizes an example of the Python equivalent, or None."""
	if not is_portable(t):
		raise MessageRegistrationError(None, 'non-memory type presented for construction - %r' % (t,))

	try:
		c = FAKE_CLASS[t.__class__]
		return c()
	except KeyError:
		pass
	except AttributeError:
		raise MessageRegistrationError(None, 'internal failure to create from class')

	# Following types are more involved - cant be
	# ctor'd solely from the class.
	if isinstance(t, Enumeration):
		e = t.element
		s = e._member_names_[0]
		f = e[s]
		return f
	elif isinstance(t, VectorOf):
		e = fake(t.element)
		d = [e]
		return d
	elif isinstance(t, ArrayOf):
		d = [None] * t.size
		for i in range(t.size):
			d[i] = fake(t.element)
		return d
	elif isinstance(t, DequeOf):
		e = fake(t.element)
		d = deque()
		d.append(e)
		return d
	elif isinstance(t, SetOf):
		e = fake(t.element)
		d = set()
		d.add(e)
		return d
	elif isinstance(t, MapOf):
		k = fake(t.key)
		v = fake(t.value)
		d = {}
		d[k] = v
		return d
	elif isinstance(t, UserDefined):
		element = t.element
		schema = element.__art__.schema
		d = element()
		for k, v in schema.items():
			setattr(d, k, fake(v))
		return d
	elif isinstance(t, PointerTo):
		return fake(t.element)

	raise MessageRegistrationError(None, 'internal failure to create from memory')



# test_function.py
import random
import kipjak as kj

random.seed()


def texture(self, x: int=8, y: int=8) -> list[list[float]]:
	'''Generate a matrix and fill with random values. Return a 2D table of floats.'''

	table = []
	for r in range(y):
		row = [None] * x
		table.append(row)
		for c in range(x):
			row[c] = random.random()	# Populate table.

	return table

kj.bind(texture)			# Register with framework.

if __name__ == '__main__':	# Process entry-point.
	kj.create(texture)



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
__docformat__ = 'restructuredtext'

import errno
import platform
import queue
import threading
import socket
import select
import re
import uuid
from nacl.public import PrivateKey, PublicKey, Box
from enum import Enum
from datetime import datetime

from .general_purpose import *
from .ip_networking import *
from .object_space import *
from .convert_memory import *
from .virtual_memory import *
from .message_memory import *
from .convert_type import *
from .virtual_codec import *
from .json_codec import *
from .virtual_runtime import *
from .virtual_point import *
from .point_runtime import *
from .point_machine import *
from .object_runtime import *
from .bind_type import *
from .http import ApiServerStream, ApiClientSession, ApiClientStream, ReForm

__all__ = [
	'ListenForStream',
	'ConnectStream',
	'Listening',
	'Accepted',
	'Connected',
	'NotListening',
	'NotAccepted',
	'NotConnected',
	'EndOfTransport',
	'Close',
	'Closed',
	'listen',
	'connect',
	'stop_listening',
]

# Platform dependent operation. Initially - windows doesnt
# like partial socket close.
PLATFORM_SYSTEM = platform.system()

if PLATFORM_SYSTEM == 'Windows':
	SHUT_SOCKET = socket.SHUT_RDWR
else:
	SHUT_SOCKET = socket.SHUT_RD

TS = Gas(sockets=None, channel=None)

# Machine states.
class INITIAL: pass
class PENDING: pass
class NORMAL: pass
class CHECKING: pass
class CLEARING: pass
class PAUSING: pass

# Control messages sent to the sockets thread
# via the control channel.
class ListenForStream(object):
	"""Session request, open a server presence.

	:param lid: unique id for this listen request
	:param requested_ipp: IP and port requested by the application
	:param encrypted: enable encryption
	:param http_server: list of classes
	:param default_to_request: default to :class:`~.HttpRequest`
	"""
	def __init__(self, lid: UUID=None, requested_ipp: HostPort=None, encrypted: bool=False,
			http_server: list[Type]=None, uri_form: ReForm=None, default_to_request: bool=True):
		self.lid = lid
		self.requested_ipp = requested_ipp or HostPort()
		self.encrypted = encrypted
		self.http_server = http_server or []
		self.uri_form = uri_form
		self.default_to_request = default_to_request

class ConnectStream(object):
	"""
	Session request, open a transport to an address.

	:param requested_ipp: IP and port requested by the application
	:param encrypted: enable encryption
	:param keep_alive: monitor the connection
	:param http_client: inserted as the path in the request URI
	:param application_json: enable **kipjak** JSON body
	"""
	def __init__(self, requested_ipp: HostPort=None, encrypted: bool=False, keep_alive: bool=False,
			http_client: str=None, application_json: bool=False):
		self.requested_ipp = requested_ipp or HostPort()
		self.encrypted = encrypted
		self.keep_alive = keep_alive
		self.http_client = http_client
		self.application_json = application_json

class StopListening(object):
	def __init__(self, lid: UUID=None):
		self.lid = lid

# Update messages from sockets thread to app.
class Listening(object):
	"""Session notification, server presence established.

	:param request: original listen request
	:param listening_ipp: network address in use
	:param controller_address: object that requested the listen
	"""
	def __init__(self, request: ListenForStream=None,
			listening_ipp: HostPort=None,
			controller_address: Address=None):
		self.request = request
		self.listening_ipp = listening_ipp or HostPort()
		self.controller_address = controller_address or NO_SUCH_ADDRESS

class Accepted(object):
	"""Session notification, transport to client established.

	:param listening: listening server
	:param opened_ipp: network address of accepted transport
	:param proxy_address: address of remote party
	:param opened_at: moment of acceptance
	"""
	def __init__(self, listening: Listening=None,
			opened_ipp: HostPort=None,
			proxy_address: Address=None,
			opened_at: datetime=None):
		self.listening = listening or Listening()
		self.opened_ipp = opened_ipp or HostPort()
		self.proxy_address = proxy_address or NO_SUCH_ADDRESS
		self.opened_at = opened_at

class Connected(object):
	"""Session notification, transport to server established.

	:param request: original connect request
	:param opened_ipp: network address of connected transport
	:param proxy_address: address of remote party
	:param opened_at: moment of connection
	"""
	def __init__(self, request: ConnectStream=None, opened_ipp: HostPort=None,
			proxy_address: Address=None,
			opened_at: datetime=None):
		self.request = request or ConnectStream()
		self.opened_ipp = opened_ipp or HostPort()
		self.proxy_address = proxy_address or NO_SUCH_ADDRESS
		self.opened_at = opened_at

class NotListening(Faulted):
	"""Session notification, server not established.

	:param request: original listen request
	:param error_code: platform error number
	:param error_text: platform error message
	"""
	def __init__(self, request: ListenForStream=None, error_code: int=0, error_text: str=None):
		self.request = request or ListenForStream()
		requested_ipp = self.request.requested_ipp
		note = '' if not error_text else f' ({error_text})'
		if requested_ipp.host is None or requested_ipp.port is None:
			Faulted.__init__(self, f'cannot stop listen"{note}')
			return
		note = '' if not error_text else f' ({error_text})'
		Faulted.__init__(self, f'cannot listen at "{requested_ipp}"{note}', error_code=error_code)

class NotAccepted(Faulted):
	"""Session notification, transport to client not established.

	:param listening: listening server
	:param error_code: platform error number
	:param error_text: platform error message
	"""
	def __init__(self, listening: Listening=None, error_code: int=0, error_text: str=None):
		self.listening = listening or Listening()
		listening_ipp = self.listening.listening_ipp
		note = '' if not error_text else f' ({error_text})'
		Faulted.__init__(self, f'cannot accept at "{listening_ipp}"{note}', error_code=error_code)

class NotConnected(Faulted):
	"""Session notification, transport to server not established.

	:param request: original connect request
	:param error_code: platform error number
	:param error_text: platform error message
	"""
	def __init__(self, request: ConnectStream=None, error_code: int=0, error_text: str=None):
		self.request = request or ConnectStream()
		requested_ipp = self.request.requested_ipp
		note = '' if not error_text else f' ({error_text})'
		Faulted.__init__(self, f'cannot connect to "{requested_ipp}"{note}', error_code=error_code)

bind(ListenForStream)
bind(ConnectStream)
bind(StopListening)
bind(Listening, copy_before_sending=False)
bind(Accepted)
bind(Connected)
bind(NotListening, explanation=Unicode(), exit_status=Integer8())
bind(NotAccepted, explanation=Unicode(), exit_status=Integer8())
bind(NotConnected, explanation=Unicode(), exit_status=Integer8())

# Session termination messages. Handshake between app
# and sockets thread to cleanly terminate a connection.

class EndOfTransport(Enum):
	"""
	Enumeration of the reasons that a connect/listen transport might close.

	* ON_REQUEST - local process sent a :class:`~.Close`.
	* ABANDONED_BY_REMOTE - transport abandoned by network or remote.
	* WENT_STALE - transport closed by keep-alive machinery.
	* OUTBOUND_STREAMING - fault during outbound serialization.
	* INBOUND_STREAMING - fault during inbound serialization.
	"""
	ON_REQUEST=1
	ABANDONED_BY_REMOTE=2
	WENT_STALE=3
	OUTBOUND_STREAMING=4
	INBOUND_STREAMING=5

class Close(object):
	"""Session control, terminate the messaging transport.

	:param message: completion message for the session
	:type message: :ref:`message<kj-message>`
	:param reason: what prompted the termination
	:param note: short diagnostic
	:param error_code: platform error code
	"""
	def __init__(self, message: Any=None, reason: EndOfTransport=None, note: str=None, error_code: int=None):
		self.message = message
		self.reason = reason or EndOfTransport.ON_REQUEST
		self.note = note
		self.error_code = error_code

class Closed(object):
	"""
	Session notification, local termination of the messaging transport.

	:param message: completion message for the session
	:type message: :ref:`message<kj-message>`
	:param reason: what prompted the termination
	:param note: short diagnostic
	:param error_code: platform error code
	:param opened_ipp: network address of terminated transport
	:param opened_at: moment of termination
	"""
	def __init__(self, message: Any=None, reason: EndOfTransport=None,
			note: str=None, error_code: int=None,
			opened_ipp: HostPort=None, opened_at: datetime=None):
		self.message = message
		self.reason = reason or EndOfTransport.ON_REQUEST
		self.note = note
		self.error_code = error_code
		self.opened_ipp = opened_ipp or HostPort()
		self.opened_at = opened_at

bind(Close, copy_before_sending=False)
bind(Closed, copy_before_sending=False)

#
#
class Shutdown(object):
	def __init__(self, s=None, close: Close=None):
		self.s = s
		self.close = close or Close()

class Bump(object):
	def __init__(self, s=None):
		self.s = s

bind(Shutdown, not_portable=True)
bind(Bump, not_portable=True)

# Classes representing open sockets for one reason or another;
# - ControlChannel.... accepted end of backdoor into sockets loop.
# - TcpServer ........ an active listen
# - TcpClient ........ an active connect
# - TcpTransport ........ established transport, child of listen or connect

class ControlChannel(object):
	def __init__(self, s):
		self.s = s

class TcpServer(object):
	def __init__(self, s, request, listening, controller_address, named_type, search_subs):
		self.s = s
		self.request = request
		self.listening = listening
		self.controller_address = controller_address
		self.named_type = named_type
		self.search_subs = search_subs

	def encrypted(self):
		return self.request.encrypted

class TcpClient(object):
	def __init__(self, s, request, connected, controller_address):
		self.s = s
		self.request = request
		self.connected = connected
		self.controller_address = controller_address

	def encrypted(self):
		return self.request.encrypted

	def keep_alive(self):
		return self.request.keep_alive

# Underlying network constraints.
#
TCP_RECV = 4096
TCP_SEND = 4096
UDP_RECV = 4096
UDP_SEND = 4096

# Security/reliability behaviours.
#
NUMBER_OF_DIGITS = (7 * 3) + 2
GIANT_FRAME = 1048576

#
#
class Header(object):
	def __init__(self, to_address=None, return_address=None, tunnel: bool=False):
		self.to_address = to_address
		self.return_address = return_address
		self.tunnel = tunnel

bind(Header, to_address=TargetAddress(), return_address=Address())

HEADING = UserDefined(Header)
BOOK = MapOf(Unicode(),Address())

#
#
class Relay(object):
	def __init__(self, block: bytearray=None, address_book: dict[str, Address]=None):
		self.block = block
		self.address_book = address_book

bind(Relay)

# Conversion of messages to on-the-wire blocks, and back again.
# The default, fully typed, async, bidirectional messaging.
class MessageStream(object):
	def __init__(self, transport):
		self.transport = transport

		# Specific to input decoding.
		self.analysis_state = 1
		self.size_byte = bytearray()
		self.size_len = []
		self.frame_size = 0
		self.frame_byte = bytearray()

		# Inbound FSM processing of a frame.
		def s1(c):
			if c in b'0123456789,':
				nd = len(self.size_byte)
				if nd < NUMBER_OF_DIGITS:
					self.size_byte.append(c)
					return 1
				raise OverflowError(f'unlikely frame size with {nd} digits')
			elif c == 10:	# ord('\n')
				a = self.size_byte.split(b',')
				if len(a) != 3:
					raise ValueError(f'unexpected dimension')
				for b in a:
					if not b or not b.isdigit():
						raise ValueError(f'mangled frame dimensions')
				s0 = int(a[0])
				s1 = int(a[1])
				s2 = int(a[2])
				self.size_len = [s0, s1, s2]
				if s0 > s2 or (s0 + s1) > s2:
					raise ValueError(f'unlikely frame offsets')
				self.jump_size = s2
				if self.jump_size > GIANT_FRAME:
					raise OverflowError(f'oversize frame of {self.jump_size} bytes')
				elif self.jump_size == 0:
					return 3
				return 2
			raise ValueError(f'frame with unexpected {c} in digits')

		def s2(c):
			self.frame_byte.append(c)
			self.jump_size -= 1
			if self.jump_size == 0:
				return 3
			return 2

		def s3(c):
			if c == 10:
				return 0
			raise ValueError(f'unexpected {c} at end-of-frame')

		self.shift = {
			1: s1,
			2: s2,
			3: s3,
		}

	# Push a message onto the byte stream.
	def message_to_block(self, mtr):
		m, t, r = mtr
		encoded_bytes = self.transport.encoded_bytes
		tunnel = False
		key_box = self.transport.key_box
		codec = self.transport.codec

		# Types significant to streaming.
		# Be nice to move DH detection elsewhere.
		if isinstance(m, tuple) and isinstance(m[1], Block):
			if m[0] is not None:
				tunnel = True
		elif isinstance(m, Diffie):
			self.transport.private_key = PrivateKey.generate()
			public_bytes = self.transport.private_key.public_key.encode()
			m.public_key = bytearray(public_bytes)
		elif isinstance(m, Hellman):
			self.transport.private_key = PrivateKey.generate()
			shared_bytes = bytes(m.public_key)
			shared_key = PublicKey(shared_bytes)
			self.transport.key_box = Box(self.transport.private_key, shared_key)
			public_bytes = self.transport.private_key.public_key.encode()
			m.public_key = bytearray(public_bytes)

		# Bring the parts together.
		# 1. Header
		h = Header(t, r, tunnel)
		e = codec.encode(h, HEADING)
		b0 = e.encode('utf-8')
		n0 = len(b0)

		# 2. Message body - 1 of following 3.
		if tunnel:
			# b1 = m.block
			b1 = m[0]
			n1 = len(b1)
			s = codec.encode({}, BOOK)
		elif isinstance(m, Relay):
			b1 = m.block
			n1 = len(b1)
			s = codec.encode(m.address_book, BOOK)
		else:
			address_book = {}
			e = codec.encode(m, Any(), address_book=address_book)
			b1 = e.encode('utf-8')
			n1 = len(b1)
			s = codec.encode(address_book, BOOK)

		# 3. Mutated addresses.
		b2 = s.encode('utf-8')

		# Combine into 1 and optionally encrypt.
		b0 += b1
		b0 += b2
		if key_box:
			b0 = key_box.encrypt(b0)
		n3 = len(b0)

		# Put frame on the transport.
		n = f'{n0},{n1},{n3}'
		encoded_bytes += n.encode('ascii')
		encoded_bytes += b'\n'
		encoded_bytes += b0
		encoded_bytes += b'\n'

	# Complete zero or more messages, using the given block.
	def recover_message(self, received, sockets):
		# Need a loop here because of encryption handshaking.
		codec = self.transport.codec
		proxy_address = self.transport.proxy_address
		diffie_hellman = self.transport.diffie_hellman

		for h, b_, a in self.recover_frame(received):
			s = h.decode('utf-8')
			header = codec.decode(s, HEADING)
			s = a.decode('utf-8')
			address_book = codec.decode(s, BOOK)

			to_address = header.to_address
			return_address = header.return_address

			if header.tunnel:					# Binary block - directly from the frame.
				body = bytearray(b_)
				body = cast_to(body, bytearray_type)

			elif len(header.to_address) > 1:	# Passing through. Just received and headed back out.
				body = Relay(b_, address_book)

			else:
				# Need to recover the fully-typed message.
				s = b_.decode('utf-8')
				body = codec.decode(s, Any(), address_book=address_book)

				#for address, path in space:
				#	poke(body, address, path)

				# Handling of encryption handshaking and keep-alives.
				if isinstance(body, Diffie):
					sockets.send(Hellman(body.public_key), proxy_address)
					if not diffie_hellman:
						continue
					h = diffie_hellman[0]
					body, to_address, return_address = h
				elif isinstance(body, Hellman):
					public_bytes = bytes(body.public_key)
					public_key = PublicKey(public_bytes)
					self.transport.key_box = Box(self.transport.private_key, public_key)
					if not diffie_hellman:
						continue
					h = diffie_hellman[0]
					body, to_address, return_address = h
				elif isinstance(body, KeepAlive):
					sockets.send(OpenKeep(body, return_address), proxy_address)
					continue
				else:
					pass	# Normal application messaging.

			yield body, to_address, return_address

	# Pull zero or more frames from the given block.
	def recover_frame(self, received):
		key_box = self.transport.key_box
		for c in received:
			next = self.shift[self.analysis_state](c)
			if next:
				self.analysis_state = next
				continue

			# Completed frame.
			f = bytes(self.frame_byte)
			if key_box:
				f = key_box.decrypt(f)

			# Breakout parts and yield.
			n0 = self.size_len[0]
			n1 = self.size_len[1]
			b2 = n0 + n1
			yield f[0:n0], f[n0:b2], f[b2:]

			# Restart.
			self.analysis_state = 1
			self.size_byte = bytearray()
			self.size_len = []
			self.frame_size = 0
			self.frame_byte = bytearray()

# Generic section of all network messaging.
class TcpTransport(object):
	def __init__(self, messaging_type, parent, controller_address, opened):
		self.messaging = messaging_type(self)
		self.parent = parent
		self.controller_address = controller_address
		self.return_proxy = None
		self.local_termination = None
		self.proxy_address = None

		self.codec = None

		self.pending = []			# Messages not yet in the loop.
		self.lock = threading.RLock()		# Safe sharing and empty detection.
		self.messages_to_encode = deque()

		self.encoded_bytes = bytearray()

		self.diffie_hellman = None
		self.private_key = None
		self.key_box = None

		self.opened = opened
		self.closing = None

	def set_routing(self, return_proxy, local_termination, proxy_address):
		# Define addresses for message forwarding.
		# return_proxy ........ address that response should go back to.
		# local_termination ... address of default target, actor or session.
		# proxy_address ...... source address of connection updates, session or proxy.
		self.codec = CodecJson(return_proxy=return_proxy, local_termination=local_termination)
		self.return_proxy = return_proxy
		self.local_termination = local_termination
		self.proxy_address = proxy_address

	# Output
	# Application to proxy.
	def put(self, m, t, r):
		try:
			self.lock.acquire()
			empty = len(self.pending) == 0
			t3 = (m, t, r)
			self.pending.append(t3)
		finally:
			self.lock.release()
		return empty

	def drain(self, a):
		try:
			self.lock.acquire()
			count = len(self.pending)
			a.extend(self.pending)
			self.pending = []
		finally:
			self.lock.release()
		return count

	# Proxy to transport.
	def send_a_block(self, s):
		t = self.queue_to_block()
		if t == 0:
			return False
		n = t if t <= TCP_SEND else TCP_SEND
		chunk = self.encoded_bytes[:n]
		n = s.send(chunk)
		if n:
			self.encoded_bytes = self.encoded_bytes[n:]
			return True
		return False

	def queue_to_block(self):
		encoded_bytes = self.encoded_bytes
		while len(encoded_bytes) < TCP_SEND:
			if len(self.messages_to_encode) == 0:
				added = self.drain(self.messages_to_encode)
				if added == 0:
					break
			# The message and to-return addresses.
			mtr = self.messages_to_encode.popleft()
			self.messaging.message_to_block(mtr)

		# Bytes available for a send.
		return len(encoded_bytes)

	# Input.
	def receive_a_message(self, received, sockets):
		for body, to_address, return_address in self.messaging.recover_message(received, sockets):
			sockets.forward(body, to_address, return_address)

#
#
class KeepAlive(object):
	def __init__(self, seconds: float=0.0):
		self.seconds = seconds

class OpenKeep(object):
	def __init__(self, keep: KeepAlive=None, return_address: Address=None):
		self.keep = keep
		self.return_address = return_address

class StillThere(object):
	def __init__(self, seconds: float=0.0):
		self.seconds = seconds

bind(KeepAlive, copy_before_sending=False, execution_trace=False, message_trail=False)
bind(StillThere, copy_before_sending=False, execution_trace=False, message_trail=False)
bind(OpenKeep, copy_before_sending=False, execution_trace=False, message_trail=False)

IDLE_TRANSPORT = 60.0
RESPONSIVE_TRANSPORT = 5.0

DELAYED_KEEPALIVE = T1
REQUESTED_CHECK = T2

class SocketKeeper(Point, StateMachine):
	def __init__(self, proxy_address=None, expecting=None):
		Point.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.proxy_address = proxy_address
		self.expecting = expecting

		self.log_count = 0

	def first_few(self):
		self.log_count += 1
		if self.log_count < 4:
			return True
		return False

	def originate(self, message, address):
		s = spread_out(IDLE_TRANSPORT, 5)
		self.send(message(s), address)
		return s

def SocketKeeper_INITIAL_Start(self, message):
	if self.expecting is None:
		seconds = self.originate(KeepAlive, self.proxy_address)
		self.start(T3, seconds + 3.0)						# Expect response.
		self.log(USER_TAG.TRACE, f'Keeper enabled at local client (requests remote silence {seconds:.1f})')
		return CHECKING

	self.log(USER_TAG.TRACE, f'Keeper enabled by remote client (observing requested silence {self.expecting:.1f})')
	self.start(T2, self.expecting)
	return PENDING

def SocketKeeper_PAUSING_T1(self, message):
	# DEFUNCT
	seconds = self.originate(KeepAlive, self.proxy_address)
	self.start(T3, seconds + 3.0)						# Expect response.
	self.log(USER_TAG.TRACE, f'Client enables server (requests radio silence {seconds:.1f})')
	return CHECKING

def SocketKeeper_PAUSING_Stop(self, message):
	self.complete(Aborted())

def SocketKeeper_PENDING_T2(self, message):		# Fulfil expectation and start own cycle.
	seconds = self.originate(StillThere, self.proxy_address)
	self.start(T3, seconds + 5.0)						# Expect response.
	if self.first_few():
		self.log(USER_TAG.TRACE, f'Silence observed, requests remote silence ({seconds:.1f})')
	return CHECKING

def SocketKeeper_PENDING_Stop(self, message):
	self.complete(Aborted())

def SocketKeeper_CHECKING_StillThere(self, message):
	self.proxy_address = self.return_address
	self.cancel(T3)
	self.start(T2, message.seconds)						# All good. Keep going.
	if self.first_few():
		self.log(USER_TAG.TRACE, f'Silence honoured by remote (observing requested silence {message.seconds:.1f})')
	return PENDING

def SocketKeeper_CHECKING_T3(self, message):
	self.log(USER_TAG.WARNING, f'Silence request timed out, closing')
	self.send(Close(reason=EndOfTransport.WENT_STALE, note='unresponsive'), self.proxy_address)
	self.complete(TimedOut(T3))

def SocketKeeper_CHECKING_Stop(self, message):
	self.complete(Aborted())

SOCKET_KEEPER_DISPATCH = {
	INITIAL: (
		(Start,), ()
	),
	PAUSING: (
		(T1, Stop), ()
	),
	PENDING: (
		(T2, Stop), ()
	),
	CHECKING: (
		(StillThere, T3, Stop), ()
	),
}

bind(SocketKeeper, SOCKET_KEEPER_DISPATCH, thread='keeper', execution_trace=False, message_trail=False)

#
#
class SocketProxy(Point, StateMachine):
	"""Local representation of an object at remote end of a network connection.

	:param s: associated network connection
	:type s: socket descriptor
	:param channel: async socket loop
	:type channel: internal
	:param transport: associated buffering
	:type transport: internal
	"""
	def __init__(self, s, channel, transport, keep_alive=False):
		Point.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.s = s
		self.channel = channel
		self.transport = transport
		self.keep_alive = keep_alive
		self.keeper = None
		self.checked = 0

	def first_few(self):
		if self.checked > 4:
			return False
		self.checked += 1
		return True

SOCKET_DOWN = (errno.ECONNRESET, errno.EHOSTDOWN, errno.ENETDOWN, errno.ENETRESET)

def SocketProxy_INITIAL_Start(self, message):
	if self.keep_alive:
		self.keeper = self.create(SocketKeeper, proxy_address=self.object_address)
	return NORMAL

def SocketProxy_NORMAL_OpenKeep(self, message):
	self.keeper = self.create(SocketKeeper, proxy_address=message.return_address, expecting=message.keep.seconds)
	return NORMAL

def SocketProxy_NORMAL_Unknown(self, message):
	message = cast_to(message, self.received_type)
	empty = self.transport.put(message, self.to_address, self.return_address)
	if empty:
		self.channel.send(Bump(self.s), self.object_address)
	return NORMAL

def SocketProxy_NORMAL_Close(self, message):
	self.channel.send(Shutdown(self.s, message), self.object_address)
	if self.keeper:
		self.send(Stop(), self.keeper)
		return CLEARING
	self.complete()

def SocketProxy_NORMAL_Stop(self, message):
	# Close created internally by ListenConnect.
	self.channel.send(Shutdown(self.s), self.object_address)
	if self.keeper:
		self.send(Stop(), self.keeper)
		return CLEARING
	self.complete()

def SocketProxy_CLEARING_Returned(self, message):
	self.complete()

TCP_PROXY_DISPATCH = {
	INITIAL: (
		(Start,),
		()
	),
	NORMAL: (
		(OpenKeep, Unknown, Close, Stop),
		()
	),
	CLEARING: (
		(Returned,),
		()
	),
}

bind(SocketProxy, TCP_PROXY_DISPATCH, thread='socketry')

#
#
class Diffie(object):
	def __init__(self, public_key: bytearray=None):
		self.public_key = public_key

class Hellman(object):
	def __init__(self, public_key: bytearray=None):
		self.public_key = public_key

bind(Diffie, copy_before_sending=False)
bind(Hellman, copy_before_sending=False)

# Signals from the network represented
# as distinct classes - for dispatching.
class ReceiveBlock: pass
class ReadyToSend: pass
class BrokenTransport: pass

# CONTROL CHANNEL
# First two functions are for handling the 1-byte events
# coming across the control socket.
def ControlChannel_ReceiveBlock(self, control, s):
	s.recv(1)					   # Consume the bump.
	mr = self.pending.get()

	# This second jump is to simulate the common handling of control
	# channel events and select events.
	c = type(mr[0])
	try:
		f = SELECT_TABLE[(ControlChannel, c)]
	except KeyError:
		self.warning(f'unknown message received on control channel ({c})')
		return
	f(self, control, mr)

def ControlChannel_BrokenTransport(self, control, s):
	self.fault('control channel broken')
	self.clear_out(s)

# The rest of them handle the simulated receive of the
# actual message.
def ControlChannel_ListenForStream(self, control, mr):
	m, r = mr
	requested_ipp = m.requested_ipp

	named_type = None
	search_subs = None
	if m.http_server:
		named_type = {t.__art__.name: t for t in m.http_server}
	elif m.uri_form:
		if m.uri_form.form is None:
			error_text = 'URI provided with no pattern'
			nl = NotListening(m, error_code=e.errno, error_text=error_text)
			self.send(nl, r)
			return

		try:
			search_subs = m.uri_form.compile_form()
		except re.error as e:
			error_text = f'URI pattern does not compile, {e}'
			self.trace(f'{error_text} "{m.uri_form.form}"')
			nl = NotListening(m, error_text=error_text)
			self.send(nl, r)
			return

	if not self.running:
		nl = NotListening(m, error_code=0, error_text='sockets shutting down')
		self.send(nl, r)
		return

	try:
		server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	except socket.error as e:
		nl = NotListening(m, error_code=e.errno, error_text=str(e))
		self.send(nl, r)
		return

	def server_not_listening(e):
		server.close()
		nl = NotListening(m, error_code=e.errno, error_text=str(e))
		self.send(nl, r)

	try:
		server.setblocking(False)
		server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		server.bind(requested_ipp.inet())
		server.listen(5)
	except (socket.herror, socket.gaierror, socket.error) as e:
		server_not_listening(e)
		return
	except OverflowError as e:
		server.close()
		nl = NotListening(m, error_code=0, error_text=str(e))
		self.send(nl, r)
		return

	hap = server.getsockname()
	listening_ipp = HostPort(hap[0], hap[1])

	note = ' (encrypted)' if m.encrypted else ''
	self.trace(f'Listening{note} on "{listening_ipp}", requested "{requested_ipp}"')

	listening = Listening(m, listening_ipp=listening_ipp, controller_address=r)

	self.networking[server] = TcpServer(server, m, listening, r, named_type, search_subs)
	self.receiving.append(server)
	self.faulting.append(server)

	self.lid[m.lid] = server

	self.send(listening, r)

def close_ending(proxy):
	def ending(message, parent, address):
		send_a_message(Close(message), proxy, address)
	return ending

def open_stream(self, parent, s, opened):
	controller_address = parent.controller_address
	keep_alive = False
	ts = MessageStream

	if isinstance(parent, TcpClient):
		keep_alive = parent.keep_alive()
		if parent.request.http_client:
			ts = ApiClientStream
	elif isinstance(parent, TcpServer):
		if len(parent.request.http_server) > 0 or parent.request.uri_form:
			ts = ApiServerStream

	transport = TcpTransport(ts, parent, controller_address, opened)
	proxy_address = self.create(SocketProxy, s, self.channel, transport, keep_alive=keep_alive, object_ending=no_ending)

	if ts == ApiClientStream:
		ending = close_ending(proxy_address)
		session_address = self.create(ApiClientSession,
			controller_address=controller_address, proxy_address=proxy_address,
			object_ending=ending)
		transport.set_routing(proxy_address, session_address, session_address)
	else:
		transport.set_routing(proxy_address, controller_address, proxy_address)

	self.networking[s] = transport
	return transport, proxy_address

def ControlChannel_ConnectStream(self, control, mr):
	m, r = mr
	requested_ipp = m.requested_ipp

	if not self.running:
		nc = NotConnected(m, 0, 'sockets shutting down')
		self.send(nc, r)
		return

	try:
		client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		client.setblocking(False)
	except socket.error as e:
		self.send(NotConnected(m, e.errno, str(e)), r)
		return

	def client_not_connected(e):
		client.close()
		self.send(NotConnected(m, e.errno, str(e)), r)

	try:
		e = client.connect_ex(requested_ipp.inet())
		if e:
			# Connect request cannot complete. Check for codes indicating
			# async issue. If not it's a real error.
			if e not in (errno.EINPROGRESS, errno.EWOULDBLOCK, errno.EAGAIN):
				client.close()
				self.send(NotConnected(m, e, 'Connect incomplete and no pending indication.'), r)
				return

			# Build a transient "session" that just exists to catch
			# an initial, either send or fault (a receive is treated
			# as an error). True session is constructed on receiving
			# a "normal" send event.
			pending = TcpClient(client, m, None, r)

			self.networking[client] = pending
			self.receiving.append(client)
			self.sending.append(client)
			self.faulting.append(client)
			return

	except (socket.herror, socket.gaierror, socket.error) as e:
		client_not_connected(e)
		return
	except OverflowError as e:
		client.close()
		self.send(NotConnected(m, 0, str(e)), r)
		return

	hap = client.getsockname()
	opened_ipp = HostPort(hap[0], hap[1])
	connected = Connected(m, opened_ipp=opened_ipp, opened_at=world_now())

	parent = TcpClient(client, m, connected, r)
	transport, proxy_address = open_stream(self, parent, client, connected)
	connected.proxy_address = proxy_address

	self.networking[client] = transport
	self.receiving.append(client)
	self.sending.append(client)
	self.faulting.append(client)

	if m.encrypted:
		self.trace(f'Connected (encrypted) to "{requested_ipp}", at local address "{opened_ipp}"')
		not_connected = NotConnected(m, None, None)
		transport.diffie_hellman = (
			(connected, r, transport.proxy_address),
			(not_connected, r))
		# Start the exchange. Public key filled out during
		# streaming.
		self.send(Diffie(), proxy_address)
		return
	self.trace(f'Connected to "{requested_ipp}", at local address ""{opened_ipp}"')

	self.forward(connected, r, transport.proxy_address)

def ControlChannel_StopListening(self, control, mr):
	m, r = mr

	if not self.running:
		nl = NotListening(ListenForStream(lid=m.lid), 0, 'sockets shutting down')
		self.send(nl, r)
		return

	server = self.lid.get(m.lid, None)
	if server is None:
		e = f'no such entry "{m.lid}"'
		self.warning(e)
		not_listening = NotListening(ListenForStream(lid=m.lid), 0, e)
		self.send(not_listening, r)
		return
	server.shutdown(socket.SHUT_RDWR)

def ControlChannel_Stop(self, control, mr):
	m, r = mr

	if not self.running:
		return

	# Initiate the teardown of all application sockets.
	for k, v in self.networking.items():
		if isinstance(v, TcpTransport):
			self.send(Stop(), v.proxy_address)			# Take out proxy as well.
			#k.shutdown(socket.SHUT_RDWR)				# Direct to socket.
		elif isinstance(v, (TcpServer, TcpClient)):
			k.shutdown(socket.SHUT_RDWR)				# Direct to socket.

	# Mark the ListenConnect state. When the
	# socket count drops to 0, the instance
	# will return.
	self.running = False

def ControlChannel_Bump(self, control, mr):
	m, r = mr
	if m.s.fileno() < 0:
		# Catches the situation where the socket has been abandoned
		# by the remote and the notification to the proxy arrives behind
		# the bump.
		return
	try:
		self.sending.index(m.s)
		return
	except ValueError:
		pass
	self.sending.append(m.s)

def ControlChannel_Shutdown(self, control, mr):
	m, r = mr
	try:
		transport = self.networking[m.s]
	except KeyError:
		# Already cleared by Abandoned codepath.
		return
	transport.closing = m.close
	m.s.shutdown(SHUT_SOCKET)

# Dispatch of socket signals;
# - ReceiveBlock ...... there are bytes to recv
# - ReadyToSend ....... an opportunity to send
# - BrokenTransport ... error on socket
# and server/client/connection;
# - TcpServer ......... listen waiting to accept
# - TcpClient ......... partial connect
# - TcpTransport ......... established connection

def TcpServer_ReceiveBlock(self, server, s):
	listening = server.listening
	request = server.request

	try:
		accepted, hap = s.accept()
		accepted.setblocking(False)
	except socket.error as e:
		#if e.errno == 22:
		self.clear_out(s)
		not_listening = NotListening(request, e.errno, str(e))
		self.send(not_listening, server.controller_address)
		return
		#not_accepted = NotAccepted(listening, e.errno, str(e))
		#self.send(not_accepted, server.controller_address)
		#return

	if not self.running:
		# Do not add to operational tables, i.e. networking[].
		accepted.shutdown(socket.SHUT_RDWR)
		accepted.close()
		return

	transport, proxy_address = open_stream(self, server, accepted, None)
	self.receiving.append(accepted)
	self.sending.append(accepted)
	self.faulting.append(accepted)

	opened_ipp = HostPort(hap[0], hap[1])

	opened_at = world_now()
	accepted = Accepted(listening=listening,
		opened_ipp=opened_ipp, proxy_address=transport.proxy_address,
		opened_at=opened_at)
	transport.opened = accepted

	if request.encrypted:
		self.trace(f'Accepted (encrypted) "{opened_ipp}", listening "{listening.listening_ipp}"')
		not_accepted = NotAccepted(listening, None, None)
		transport.diffie_hellman = (
			(accepted, server.controller_address, transport.proxy_address),
			(not_accepted, server.controller_address))
		return
	self.trace(f'Accepted "{opened_ipp}", listening "{listening.listening_ipp}"')

	self.forward(accepted, server.controller_address, transport.proxy_address)

def TcpServer_BrokenTransport(self, server, s):
	listening = server.listening
	self.send(NotListening(listening.listening_ipp, 0, "signaled by networking subsystem"), server.controller_address)
	self.clear_out(s, TcpServer)

#
def TcpClient_ReceiveBlock(self, selector, s):
	client = s
	request = selector.request

	hap = client.getsockname()
	opened_ipp = HostPort(hap[0], hap[1])
	requested_ipp = request.requested_ipp

	try:
		scrap = s.recv(TCP_RECV)

		# No exception. New transport.
		connected = Connected(request, opened_ipp=opened_ipp, opened_at=world_now())

		selector.connected = connected
		transport, proxy_address = open_stream(self, selector, client, connected)
		connected.proxy_address = proxy_address

		self.trace(f'Connected to "{requested_ipp}", at local address "{opened_ipp}"')

		if selector.encrypted():
			self.trace(f'Connected (encrypted) to "{requested_ipp}", at local address "{opened_ipp}"')
			not_connected = NotConnected(request, None, None)
			transport.diffie_hellman = (
				(connected, transport.controller_address, transport.proxy_address),
				(not_connected, selector.controller_address))
			self.send(Diffie(), proxy_address)
			return
		self.trace(f'Connected to "{requested_ipp}", at local address "{opened_ipp}"')

		self.forward(connected, transport.controller_address, transport.proxy_address)

		if not scrap:
			return

		try:
			transport.receive_a_message(scrap, self)
		except (CodecError, OverflowError, ValueError) as e:
			self.warning(f'cannot receive_a_message ({e})')
			c = Close(message=None, reason=EndOfTransport.INBOUND_STREAMING, note=str(e))
			close_by_socket(transport, c, s)
		return

	except socket.error as e:
		self.send(NotConnected(request, e.errno, str(e)), selector.controller_address)
		self.clear_out(s, TcpClient)
		return

def TcpClient_ReadyToSend(self, selector, s):
	client = s
	request = selector.request

	hap = client.getsockname()
	opened_ipp = HostPort(hap[0], hap[1])
	requested_ipp = request.requested_ipp

	connected = Connected(request, opened_ipp=opened_ipp, opened_at=world_now())
	selector.connected = connected

	transport, proxy_address = open_stream(self, selector, client, connected)
	connected.proxy_address = proxy_address

	if selector.encrypted():
		self.trace(f'Connected (encrypted) to "{requested_ipp}", at local address "{opened_ipp}"')
		not_connected = NotConnected(request, None, None)
		transport.diffie_hellman = (
			(connected, transport.controller_address, transport.proxy_address),
			(not_connected, selector.controller_address))
		# Start the exchange of public keys.
		self.send(Diffie(), proxy_address)
		return
	self.trace(f'Connected to "{requested_ipp}", at local address "{opened_ipp}"')

	self.forward(connected, transport.controller_address, transport.proxy_address)

def TcpClient_BrokenTransport(self, selector, s):
	request = selector.request
	requested_ipp = request.requested_ipp

	text = 'fault on pending connect, unreachable, no service at that address or blocked'
	self.send(NotConnected(request, 0, text), selector.controller_address)
	self.clear_out(s, TcpClient)

#
def close_by_socket(transport, close, s):
	transport.closing = close
	s.shutdown(SHUT_SOCKET)

def clear_out_session(self, transport, s, reason=None, note=None, error_code=None):
	ipp = transport.opened.opened_ipp

	if transport.closing:
		close = transport.closing
		c = Closed(message=close.message,
			reason=close.reason,
			note=close.note,
			error_code=close.error_code,
			opened_ipp=ipp,
			opened_at=transport.opened.opened_at)
	else:
		self.send(Stop(), transport.proxy_address)
		c = Closed(reason=reason,
			note=note,
			error_code=error_code,
			opened_ipp=ipp,
			opened_at=transport.opened.opened_at)

	self.forward(c, transport.controller_address, transport.proxy_address)
	self.clear_out(s, TcpTransport)

def TcpTransport_ReadyToSend(self, transport, s):
	try:
		if transport.send_a_block(s):
			return
	except (CodecError, OverflowError, ValueError) as e:
		self.warning(f'cannot send_a_block ({e})')
		c = Close(message=None, reason=EndOfTransport.OUTBOUND_STREAMING, note=str(e))
		close_by_socket(transport, c, s)
		return

	# Had nothing to send.
	try:
		self.sending.remove(s)
	except ValueError:
		pass

# A network transport for the purpose of exchanging
# messages between machines.

def TcpTransport_ReceiveBlock(self, transport, s):
	try:
		scrap = s.recv(TCP_RECV)
		if not scrap:
			clear_out_session(self, transport, s, reason=EndOfTransport.ABANDONED_BY_REMOTE)
			return

		try:
			transport.receive_a_message(scrap, self)
		except (CodecError, OverflowError, ValueError) as e:
			self.warning(f'Cannot receive_a_message ({e})')
			c = Close(message=None, reason=EndOfTransport.INBOUND_STREAMING, note=str(e))
			close_by_socket(transport, c, s)
		return

	except socket.error as e:
		# if e.errno == errno.ECONNREFUSED:
		#	self.warning('Connection refused')
		# elif e.errno not in SOCKET_DOWN:
		#	self.fault(f'Socket termination ({e})')
		clear_out_session(self, transport, s, reason=EndOfTransport.INBOUND_STREAMING, note=str(e), error_code=e.errno)
		return

def TcpTransport_BrokenTransport(self, selector, s):
	clear_out_session(self, selector, s, reason=EndOfTransport.INBOUND_STREAMING, note='broken transport')

#
def control_channel():
	'''Create a listen-connect pair for control channel to engine.'''
	server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	server.setblocking(False)
	server.bind(("127.0.0.1", 0))
	server.listen(1)

	server_address = server.getsockname()

	client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	client.setblocking(False)
	e = client.connect_ex(server_address)

	readable, writable, exceptional = select.select([server], [], [server])
	if not readable:
		client.close()
		server.close()
		raise RuntimeError('Forming control channel, select has not received connect notification.')

	accepted, client_address = server.accept()
	accepted.setblocking(False)

	accept_address = accepted.getsockname()

	return server, accepted, client

def control_close(lac):
	# Close the listen and the connect. Accepted
	# will be closed by ListenConnect.
	lac[2].close()
	lac[0].close()

#
#
BUMP = b'X'

class SocketChannel(object):
	def __init__(self, pending=None, client=None):
		'''
		This is the per-object client end of the control
		channel into the network I/O loop.
		'''
		self.pending = pending
		self.client = client

	def send(self, message, address):
		self.pending.put((message, address))
		buffered = self.client.send(BUMP)
		if buffered != 1:
			raise RuntimeError('Control channel not accepting commands.')

# Damn. Sent from sockets thread to creator. They
# need it to inject messages into loop.
bind(SocketChannel, not_portable=True, copy_before_sending=False)

#
SELECT_TABLE = {
	# Handling of inbound control messages.
	(ControlChannel, ReceiveBlock):		ControlChannel_ReceiveBlock,		# Signals down the control channel.
	(ControlChannel, BrokenTransport):	ControlChannel_BrokenTransport,

	# Made to look as if the select thread can actually receive
	# sockets signals and application messages. Called from above.
	(ControlChannel, ListenForStream):	ControlChannel_ListenForStream,		# Process signals to sockets.
	(ControlChannel, ConnectStream):	ControlChannel_ConnectStream,
	(ControlChannel, Shutdown):			ControlChannel_Shutdown,
	(ControlChannel, Bump):			 	ControlChannel_Bump,
	(ControlChannel, StopListening):	ControlChannel_StopListening,
	(ControlChannel, Stop):		  		ControlChannel_Stop,

	# Operational sockets
	(TcpServer,	ReceiveBlock):	   		TcpServer_ReceiveBlock,			# Accept inbound connections.
	(TcpServer,	BrokenTransport):		TcpServer_BrokenTransport,

	(TcpClient, ReceiveBlock):			TcpClient_ReceiveBlock,			# Deferred connections.
	(TcpClient, ReadyToSend):		 	TcpClient_ReadyToSend,
	(TcpClient, BrokenTransport):	 	TcpClient_BrokenTransport,

	(TcpTransport, ReceiveBlock):		TcpTransport_ReceiveBlock,
	(TcpTransport, ReadyToSend):		TcpTransport_ReadyToSend,
	(TcpTransport, BrokenTransport):	TcpTransport_BrokenTransport,
}

class ListenConnect(Threaded, Stateless):
	def __init__(self):
		Threaded.__init__(self)
		Stateless.__init__(self)

		# Construct the control channel and access object.
		self.pending = queue.Queue()
		self.lac = control_channel()
		self.channel = SocketChannel(self.pending, self.lac[2])

		# Load control details into socket tables.
		self.listening = self.lac[0]
		self.accepted = self.lac[1]
		self.networking = {
			self.accepted: ControlChannel(self.accepted),	# Receives 1-byte BUMPs.
		}

		# Active socket lists for select.
		self.receiving = [self.accepted]
		self.sending = []
		self.faulting = self.receiving + self.sending

		self.lid = {}

		# Live.
		self.running = True

	def clear_out(self, s, expected=None):
		# Remove the specified socket from operations.
		try:
			t = self.networking[s]
		except KeyError:
			self.warning('Attempt to remove unknown socket')
			return None

		if expected and not isinstance(t, expected):
			self.warning(f'unexpected networking object {t} (expecting {expected})')
			return None

		def find():
			for lid, server in self.lid.items():
				if server == s:
					return lid
			return None
		f = find()
		if f is not None:
			del self.lid[f]

		del self.networking[s]
		try:
			self.receiving.remove(s)
		except ValueError:
			pass
		try:
			self.sending.remove(s)
		except ValueError:
			pass
		try:
			self.faulting.remove(s)
		except ValueError:
			pass
		s.close()
		return t

def ListenConnect_Start(self, message):
	# Provide channel details to parent for access
	# by application.
	self.send(self.channel, self.parent_address)

	while self.running or len(self.networking) > 1:
		R, S, F = select.select(self.receiving, self.sending, self.faulting)

		for r in R:
			try:
				a = self.networking[r]
				c = a.__class__
				j = SELECT_TABLE[(c, ReceiveBlock)]
			except KeyError:
				continue
			except ValueError:
				continue
			j(self, a, r)

		for s in S:
			try:
				a = self.networking[s]
				c = a.__class__
				j = SELECT_TABLE[(c, ReadyToSend)]
			except KeyError:
				continue
			except ValueError:
				continue
			j(self, a, s)

		for f in F:
			try:
				a = self.networking[f]
				c = a.__class__
				j = SELECT_TABLE[(c, BrokenTransport)]
			except KeyError:
				continue
			except ValueError:
				continue
			j(self, a, f)

	control_close(self.lac)
	self.complete(Ack())

bind(ListenConnect, (Start,))

# Managed creation of socket engine.
def create_sockets(root):
	TS.sockets = root.create(ListenConnect)
	m, i = root.select(SocketChannel)
	TS.channel = m

def stop_sockets(root):
	TS.channel.send(Stop(), root.object_address)
	root.select()

AddOn(create_sockets, stop_sockets)

# Interface to the engine.
def listen(self: Point, requested_ipp: HostPort, encrypted: bool=False,
			http_server: list[Type]=None, uri_form: ReForm=None, default_to_request: bool=True):
	"""
	Establishes a network presence at the specified IP
	address and port number. Returns UUID.

	:param self: asynchronous identity
	:param requested_ipp: network address to listen at
	:param encrypted: enable encryption
	:param http_server: enable HTTP with list of expected requests
	:param default_to_request: enable default conversion into HttpRequests
	:rtype: UUID
	"""
	lid = uuid.uuid4()
	ls = ListenForStream(lid=lid, requested_ipp=requested_ipp, encrypted=encrypted,
		http_server=http_server, uri_form=uri_form, default_to_request=default_to_request)
	TS.channel.send(ls, self.object_address)
	return lid

def connect(self: Point, requested_ipp: HostPort, encrypted: bool=False, keep_alive: bool=False,
			http_client: str=None, application_json: bool=False):
	"""
	Initiates a network connection to the specified IP
	address and port number.

	:param self: asynchronous identity
	:param requested_ipp: host and port to connect to
	:param encrypted: enable encryption
	:param keep_alive: enable keep-alives
	:param http_client: leading part of the outgoing request URI
	:param application_json: is the remote server a kipjak server
	"""
	cs = ConnectStream(requested_ipp=requested_ipp, encrypted=encrypted, keep_alive=keep_alive, http_client=http_client, application_json=application_json)
	TS.channel.send(cs, self.object_address)

def stop_listening(self: Point, lid: UUID):
	"""
	Shutdown the listen port with the specified identity.

	:param self: asynchronous identity
	:param lid: UUID assigned at time of :func:`~.listen`
	"""
	TS.channel.send(StopListening(lid), self.object_address)
